import os
from raft.tasks import Task, task


class SnobbyTask(Task):
    def __init__(self, *args, **kwargs):
        self.playbook = kwargs.pop('playbook')
        self.inventory = kwargs.get('inventory') or 'hosts'
        self.limit = kwargs.pop('limit', None)
        self.extra = {}
        for key, value in kwargs.items():
            if key not in (
                    'body',
                    'name',
                    'aliases',
                    'positional',
                    'optional',
                    'default',
                    'auto_shortflags',
                    'help',
                    'pre',
                    'post',
                    'autoprint',
                    'iterable',
                    'incrementable',
            ):
                self.extra[key] = value
        for key in self.extra:
            kwargs.pop(key)
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        ctx = args[0]
        extra = []
        limit = kwargs.pop('limit', None) or self.limit
        tags = kwargs.pop('tags', None) or ''
        inventory = kwargs.pop('inventory', None) or self.inventory
        verbose = kwargs.pop('verbose', False)
        if tags:
            tags = f'--tags {tags}'
        for key, value in self.extra.items():
            extra.append(f'-e {key}={value}')
        for key, value in kwargs.items():
            extra.append(f'-e {key}={value}')
        if verbose:
            extra.append('-vvvv')
        extra = ' '.join(extra)
        if limit:
            limit = f'--limit {limit}'
        command = (
            'ansible-playbook '
            f' -i {inventory}'
            f" {limit or ''}"
            f' {self.playbook}'
            f' {extra}'
            f' {tags}'
        )
        if os.path.isfile('./ansible.cfg') and 'ANSIBLE_CONFIG' not in os.environ:
            os.environ['ANSIBLE_CONFIG'] = './ansible.cfg'
        ctx.run(command, pty=True)



from raft.tasks import task
from ...base.utils import notice, notice_end
from ..base import AwsTask


__all__ = [
    'peer_vpc',
]


@task(klass=AwsTask)
def peer_vpc(ctx, vpc_id, peer_vpc_id, peer_owner_id, peer_region, session=None):
    """
    creates peering request from one vpc to another
    """
    notice(f'initiating peering from {vpc_id} to {peer_vpc_id}')
    ec2 = session.client('ec2')
    response = ec2.create_vpc_peering_connection(
        PeerVpcId=peer_vpc_id,
        PeerOwnerId=peer_owner_id,
        VpcId=vpc_id,
        PeerRegion=peer_region,
    )
    notice_end()
    print(response)



from raft.collection import Collection
from .subscription_tasks import subscriptions
from .resource_group_tasks import resource_groups
from .resource_group_tasks import providers
from .ldap_tasks import update_ldaps_cert


azure_tasks = Collection()
azure_tasks.add_task(subscriptions)
azure_tasks.add_task(resource_groups)
azure_tasks.add_task(providers)
azure_tasks.add_task(update_ldaps_cert)




from math import ceil
import time

from halo import Halo
from raft.tasks import task
from ..base import AwsTask, name_tag, yielder, get_tag
from ...base.utils import notice, notice_end, print_table, confirm


def aws_filter(name, *values):
    return {
        'Name': name,
        'Values': values
    }


def instance_state(instance):
    return instance['State']['Name'].lower()


@task(klass=AwsTask)
def recycle_instances(ctx, name, profile=None, session=None):
    instance_ids = []
    for x in yield_instances(session=session):
        i_name = name_tag(x)
        state = instance_state(x)
        if name in i_name and state == 'running':
            instance_ids.append(x['InstanceId'])
    response = confirm(
        f'You are about to delete the following '
        f'[{len(instance_ids)}] instances: '
        f'{", ".join(instance_ids)}')
    if response:
        ec2 = session.client('ec2')
        ec2.terminate_instances(InstanceIds=instance_ids)


def yield_instances(name=None, session=None):
    if name:
        name = name.lower()
    for data in yielder('ec2', 'describe_instances', session=session):
        if name:
            for x in data['Instances']:
                value = get_tag(x, 'Name') or ''
                value = value.lower()
                if name in value.lower():
                    yield x
                if x['InstanceId'] == name:
                    yield x
        else:
            yield from data['Instances']


def yield_instances_by_id(session, instance_ids):
    gn = yielder(
        'ec2', 'describe_instances',
        session=session, InstanceIds=instance_ids)
    for data in gn:
        yield from data['Instances']


@task(klass=AwsTask)
def instances(ctx, name=None, session=None, profile=None, **kwargs):
    """
    summary of instances with `name` in their `Name` tag."
    """
    header = [ 'id', 'name', 'type', 'private', 'public', 'status' ]
    rows = []
    matching_instances = []
    for x in yield_instances(name, session):
        if x['State']['Name'] == 'terminated':
            continue
        interfaces = x['NetworkInterfaces']
        public_ip = private_ip = None
        if interfaces:
            private_ip = interfaces[0]['PrivateIpAddress']
            public_ip = interfaces[0].get('PublicIpAddress', '')
        if not private_ip:
            private_ip = x.get('PrivateIpAddress', '')
        if not public_ip:
            public_ip = x.get('PublicIpAddress', '')
        rows.append([
            x['InstanceId'],
            get_tag(x, 'name'),
            x['InstanceType'],
            private_ip,
            public_ip,
            x['State']['Name'],
        ])
        matching_instances.append(x)
    rows.sort(key=lambda lx: lx[1])
    if not rows:
        print()
        print(f'nothing found matching {name}')
    else:
        print_table(header, rows)
    return matching_instances


def tag(key, value):
    return {
        'Key': key,
        'Value': value,
    }


def instance_by_name(ctx, name, session=None, prefix=None):
    notice(f'looking for {name}')
    matching_instances = list(yield_instances(name, session))
    matching_instances = [
        x for x in matching_instances
        if instance_state(x) != 'terminated'
    ]
    if not matching_instances:
        notice_end('nothing found, please halp')
    elif len(matching_instances) > 1:
        notice_end('multiple instances found')
        for instance in matching_instances:
            instance_id = instance['InstanceId']
            if name.lower() == name_tag(instance) or name.lower() == instance_id:
                ec2 = session.resource('ec2')
                x = ec2.Instance(instance_id)
                return x
        instances(ctx, name, session=session, prefix=prefix)
        return None
    instance = matching_instances[0]
    instance_id = instance['InstanceId']
    notice_end(instance_id)
    notice(f'loading instance {instance_id}')
    ec2 = session.resource('ec2')
    x = ec2.Instance(instance_id)
    notice_end()
    return x


def move_volume(
        ctx, from_name, to_name, volume_id=None,
        device_name='/dev/xvdf',
        session=None, prefix=None):
    """
    moves volume to a forensic instance for analysis
    """
    from_instance = instance_by_name(ctx, from_name)
    if not from_instance:
        return
    to_instance = instance_by_name(ctx, to_name)
    if not to_instance:
        return
    rg_volumes = from_instance.block_device_mappings
    if len(rg_volumes) > 1 and not volume_id:
        print(f'{from_name} has {len(rg_volumes)} volumes; '
              f'we don\'t know what to do')
        return
    volume_id = volume_id or rg_volumes[0]['Ebs']['VolumeId']
    notice('loading volume')
    ec2 = session.resource('ec2')
    volume = ec2.Volume(volume_id)
    volume.load()
    current_name = name_tag(volume)
    if not current_name:
        volume.create_tags(Tags=[tag('Name', name_tag(from_instance))])
    volume.load()
    current_name = name_tag(volume)
    notice_end(f'{current_name}')
    notice(f'detaching {volume_id}')
    volume.detach_from_instance()
    notice_end()
    with Halo('detaching volume', spinner='dots') as h:
        for x in range(20):
            time.sleep(1)
            volume.load()
            if volume.state == 'available':
                h.stop_and_persist(symbol='\N{heavy check mark}\N{vs16}', text='finished')
    volume.load()
    notice(f'attaching {volume_id} to {to_instance.instance_id}')
    volume.attach_to_instance(
        Device=device_name,
        InstanceId=to_instance.instance_id)
    notice_end()


def move_back_volume(
        ctx, from_name, to_name,
        device_name='/dev/sda1',
        session=None, prefix=None):
    """
    Puts the volume back on the original instance it came from now that
    we're done with forensics.
    """
    from_name = from_name.lower()
    from_instance = instance_by_name(
        ctx, from_name, session=session, prefix=prefix)
    if not from_instance:
        return
    to_name = to_name.lower()
    to_instance = instance_by_name(
        ctx, to_name, session=session, prefix=prefix)
    if not to_instance:
        return
    rg_volumes = from_instance.block_device_mappings
    current_name = 'not-a-real-name'
    volume, volume_id = None, None
    for x in rg_volumes:
        volume_id = x['Ebs']['VolumeId']
        notice(f'loading {volume_id}')
        ec2 = session.resource('ec2')
        volume = ec2.Volume(volume_id)
        volume.load()
        current_name = name_tag(volume)
        if current_name == to_name:
            notice_end('found!')
            break
        volume = None
        notice_end('not it')
    if current_name != to_name:
        notice_end('nothing to be done')
        return
    notice(f'detaching {volume_id}')
    volume.detach_from_instance()
    notice_end()
    with Halo('detaching volume', spinner='dots') as h:
        for x in range(20):
            time.sleep(1)
            volume.load()
            if volume.state == 'available':
                h.stop_and_persist(symbol='\N{heavy check mark}\N{vs16}', text='finished')
                break
    volume.load()
    notice(f'attaching {volume_id} to {to_instance.instance_id}')
    volume.attach_to_instance(
        Device=device_name,
        InstanceId=to_instance.instance_id)
    notice_end()
    notice('modifying instance')
    to_instance.load()
    to_instance.modify_attribute(
        Attribute='blockDeviceMapping',
        BlockDeviceMappings=[{
            'DeviceName': device_name,
            'Ebs': {
                'DeleteOnTermination': True,
            }
        }]
    )
    notice_end('done')


def rename_instance(ctx, old_name, new_name, session=None, prefix=None):
    notice(f'looking for {old_name}')
    matching_instances = list(yield_instances(old_name, session))
    if not matching_instances:
        notice_end('nothing found, please halp')
    elif len(matching_instances) > 1:
        notice_end('multiple instances found')
        instances(ctx, old_name, session=session, prefix=prefix)
    else:
        instance = matching_instances[0]
        instance_id = instance['InstanceId']
        notice_end(instance_id)
        notice(f'loading instance {instance_id}')
        ec2 = session.resource('ec2')
        x = ec2.Instance(instance_id)
        notice_end()
        # the create_tags api call will create a new tag *or update*
        # an existing tag.
        notice(f'changing name to {new_name}')
        x.create_tags(Tags=[tag('Name', new_name)])
        notice_end()


@task(klass=AwsTask, help={
    'name': 'the matching name of the instance(s) to start',
    'wait': 'specify this flag to wait for the command to finish',
    'force': 'specify this flag to force stop the instances, as '
             'opposed to a graceful shutdown',
})
def stop_instances(ctx, name, force=False, wait=True, session=None, prefix=None):
    """
    stops all instances with `name` in their `Name` tag.
    """
    matching_instances = instances(ctx, name, session=session, prefix=prefix)
    notice(f'looking for {name}')
    matching_instances = [
        x for x in matching_instances
        if instance_state(x) == 'running' ]
    notice_end(f'found {len(matching_instances)}')
    if not matching_instances:
        notice_end('no running instances to stop')
        return

    proceed = confirm(f'Stopping {len(matching_instances)} running instance(s)')
    if not proceed:
        return

    instance_ids = [ x['InstanceId'] for x in matching_instances ]
    client = session.client('ec2')
    client.stop_instances(InstanceIds=instance_ids, Force=force)
    if not wait:
        return
    with Halo(text='waiting', spinner='dots') as h:
        waiter = client.get_waiter('instance_stopped')
        waiter.config.delay = 15
        waiter.config.max_attempts = 40
        waiter.wait(InstanceIds=instance_ids)
        h.stop_and_persist(symbol='\N{octagonal sign}\N{vs16}')
        # h.succeed('\N{heavy check mark}\N{vs16}')


@task(klass=AwsTask, help={
    'name': 'the matching name of the instance(s) to start',
    'wait': 'specify this flag to wait for the command to finish'
})
def start_instances(ctx, name, wait=True, session=None, prefix=None):
    """
    starts all instances that have `name` in their `Name` tag.
    """
    matching_instances = instances(ctx, name, session=session, prefix=prefix)
    notice(f'looking for {name}')
    matching_instances = [
        x for x in matching_instances
        if instance_state(x) == 'stopped' ]
    notice_end(f'found {len(matching_instances)}')

    if not matching_instances:
        notice_end('no stopped instances to start')
        return

    proceed = confirm(f'Starting {len(matching_instances)} stopped instance(s)')
    if not proceed:
        return

    instance_ids = [ x['InstanceId'] for x in matching_instances ]
    client = session.client('ec2')
    client.start_instances(InstanceIds=instance_ids)
    if not wait:
        return
    with Halo(text='waiting', spinner='dots') as h:
        waiter = client.get_waiter('instance_running')
        waiter.config.delay = 3
        waiter.config.max_attempts = 200
        waiter.wait(InstanceIds=instance_ids)
        h.stop_and_persist(symbol='\N{snowman}\N{vs16}')


@task(klass=AwsTask, help={
    'n': 'the 1-based index of the volume we want to resize',
    'name': 'the name of the instance',
    'scale': 'the percent increase we want in the volume size, '
             'defaults to 100 (i.e., grow by 2x)',
})
def increase_volume_size(
        ctx, name, scale=100, index=0, session=None, prefix=None):
    """
    increases a volume by `scale` percent.
    """
    instance = instance_by_name(ctx, name, session=session, prefix=prefix)
    ec2 = session.resource('ec2')
    if not instance:
        return
    rg_volumes = instance.block_device_mappings
    for i, x in enumerate(rg_volumes, 1):
        notice(f'volume #{i}')
        volume_id = x['Ebs']['VolumeId']
        volume = ec2.Volume(volume_id)
        volume.load()
        x['volume'] = volume
        notice_end(f'{volume_id} / {volume.size} gb')
    if len(rg_volumes) > 1 and not index:
        print(f'{name} has {len(rg_volumes)} volumes; '
              f'we don\'t know what to do')
        return
    n = index or 1
    volume = rg_volumes[n - 1]['volume']
    volume_id = volume.volume_id
    new_size = volume.size * (1 + scale / 100)
    new_size = ceil(new_size)
    notice(f'increasing {volume_id} to {new_size} gb')
    client = session.client('ec2')
    client.modify_volume(VolumeId=volume_id, Size=new_size)
    notice_end('OK')


@task(klass=AwsTask, help={
    'name': 'the name of the instance as specified in a name tag or instance id'
})
def volumes(ctx, name, session=None, prefix=None):
    """
    lists all volumes attached to an instance
    """
    instance = instance_by_name(ctx, name, session=session, prefix=prefix)
    ec2 = session.resource('ec2')
    if not instance:
        return
    rg_volumes = instance.block_device_mappings
    header = [ 'volume_id', 'size', 'state', 'attachments ']
    rows = []
    for x in rg_volumes:
        volume_id = x['Ebs']['VolumeId']
        notice('volume')
        volume = ec2.Volume(volume_id)
        volume.load()
        notice_end(f'{volume_id} / {volume.size} gb')
        rows.append([
            volume_id, f'{volume.size}', volume.state,
            volume.attachments[0]['Device']
        ])
    print_table(header, rows)


@task(klass=AwsTask)
def eips(ctx, session=None, profile=None, **kwargs):
    """
    shows all eips in a given account
    """
    ec2 = session.client('ec2')
    addresses = ec2.describe_addresses()
    addresses = addresses['Addresses']
    header = [ 'name', 'ip_address', 'allocation_id' ]
    rows = []
    for x in addresses:
        rows.append([
            name_tag(x),
            x.get('PublicIp') or '',
            x['AllocationId'],
        ])
    print()
    print_table(header, rows)



import time

from raft import task
from ...base.utils import print_table
from ...base.utils import notice, notice_end
from ..base import AwsTask, yielder


@task(klass=AwsTask)
def rds_instances(ctx, name='', cluster=None, session=None, profile=None, **kwargs):
    """
    shows all rds instances and their endpoints
    """
    from ..base import name_tag
    from ..ec2 import yield_instances
    notice('getting instances')
    rds = session.client('rds')
    if cluster:
        instances = rds.describe_db_instances(Filters=[{
            'Name': 'db-cluster-id',
            'Values': [ cluster ],
        }])
    else:
        instances = rds.describe_db_instances()
    instances = instances['DBInstances']
    notice_end()
    name = name.lower()
    any_custom = False
    for x in instances:
        st_id = x['DBInstanceIdentifier']
        if name not in st_id.lower():
            continue
        if 'custom' in x['Engine']:
            any_custom = True
            break
    ec2_instances = {}
    if any_custom:
        notice('getting ec2 instance info')
        ec2_instances = {
            name_tag(x): x
            for x in yield_instances(name='db-', session=session)
        }
        notice_end()
    header = [ 'name', 'engine', 'endpoint', 'status', ]
    rows = []
    for x in instances:
        st_id = x['DBInstanceIdentifier']
        engine = x['Engine']
        endpoint = x.get('Endpoint')
        if name not in st_id.lower():
            continue
        rows.append([
            st_id,
            f"{engine} {x['EngineVersion']}",
            f"{endpoint['Address']}:{endpoint['Port']}" if endpoint else '',
            x['DBInstanceStatus'],
        ])
    rows.sort(key=lambda lx: lx[0])
    print_table(header, rows)
    print()
    if not any_custom:
        return
    header = [ 'name', 'id', 'ip', ]
    rows = []
    for x in instances:
        st_id = x['DBInstanceIdentifier']
        engine = x['Engine']
        if name not in st_id.lower():
            continue
        if 'custom' in engine:
            resource_id = x.get('DbiResourceId')
            if not resource_id:
                continue
            key = resource_id.lower()
            y = ec2_instances.get(key)
            if y:
                rows.append([
                    st_id,
                    y['InstanceId'],
                    y['PrivateIpAddress'],
                ])
    rows.sort(key=lambda lx: lx[0])
    print_table(header, rows)


@task(klass=AwsTask)
def rds_clusters(ctx, name='', session=None, profile=None, **kwargs):
    """
    shows all rds instances and their endpoints
    """
    rds = session.client('rds')
    clusters = rds.describe_db_clusters()
    name = name.lower()
    clusters = clusters['DBClusters']
    header = [ 'name', 'endpoint', ]
    rows = []
    for x in clusters:
        cluster_name = x['DBClusterIdentifier']
        cluster_name = cluster_name.lower()
        if name in cluster_name:
            rows.append([
                x['DBClusterIdentifier'],
                x['Endpoint'],
            ])
    rows.sort(key=lambda lx: lx[0])
    print_table(header, rows)


@task(klass=AwsTask)
def sys_password(ctx, cluster_name, session=None, profile=None, **kwargs):
    """
    shows the sys password for an custom rds for oracle rac database
    """
    notice('connecting to rds')
    rds = session.client('rds')
    clusters = rds.describe_db_clusters()
    clusters = clusters['DBClusters']
    notice_end()
    notice(cluster_name)
    cluster_name = cluster_name.lower()
    resource_id = None
    for x in clusters:
        st = x['DBClusterIdentifier'].lower()
        if cluster_name == st:
            resource_id = x['DbClusterResourceId']
            notice_end(resource_id)
            break
    if not resource_id:
        return
    notice('looking up sys password')
    secret_name = f'do-not-delete-rds-custom-{resource_id}-sys-'
    sm = session.client('secretsmanager')
    fn = yielder(sm, 'list_secrets', session, Filters=[{
        'Key': 'name',
        'Values': [ secret_name ],
    }])
    notice_end()
    for x in fn:
        name = x['Name']
        response = sm.get_secret_value(SecretId=name)
        print(response['SecretString'])


@task(klass=AwsTask)
def rds_ssh(ctx, name, session=None, profile=None, **kwargs):
    """
    sshes to a custom rds for oracle rac database
    """
    notice('connecting to rds')
    rds = session.client('rds')
    instances = rds.describe_db_instances(Filters=[{
        'Name': 'db-instance-id',
        'Values': [ name ],
    }])
    instances = instances['DBInstances']
    if instances:
        resource_id = instances[0]['DbiResourceId']
        notice_end(resource_id)
    else:
        notice_end('not found')
        return
    from convocations.aws.ec2 import instance_by_name
    from convocations.aws.ec2.ssh import download_secret
    x = instance_by_name(ctx, resource_id, session=session)
    key_name = x.key_name
    username = 'ec2-user'
    ip_address = x.private_ip_address
    notice('key name')
    notice_end(key_name)
    notice('ip address')
    notice_end(ip_address)
    with download_secret(session, x.key_name) as f:
        ssh_cmd = f'ssh -i {f.name} {username}@{ip_address}'
        notice_end(ssh_cmd)
        ctx.run(ssh_cmd, pty=True)
        return


@task(klass=AwsTask)
def download_rds_ssh_key(ctx, name, filename, session=None, **kwargs):
    """
    saves the ssh private key file to a specified filename
    """
    notice('connecting to rds')
    rds = session.client('rds')
    instances = rds.describe_db_instances(Filters=[{
        'Name': 'db-instance-id',
        'Values': [ name ],
    }])
    instances = instances['DBInstances']
    if instances:
        resource_id = instances[0]['DbiResourceId']
        notice_end(resource_id)
    else:
        notice_end('not found')
        return
    from convocations.aws.ec2 import instance_by_name
    from convocations.aws.ec2.ssh import download_secret
    x = instance_by_name(ctx, resource_id, session=session)
    key_name = x.key_name
    notice('key name')
    notice_end(key_name)
    with download_secret(session, x.key_name) as f:
        ctx.run(f'cp -f {f.name} {filename}')


@task(klass=AwsTask)
def delete_rac_cluster(ctx, name, session=None, **kwargs):
    """
    creates the staging oroms rds cluster and database instances
    """
    notice('connecting to rds')
    rds = session.client('rds')
    notice_end()
    notice('getting instances')
    instances = rds.describe_db_instances(Filters=[{
        'Name': 'db-cluster-id',
        'Values': [ name ]
    }])
    instances = instances['DBInstances']
    notice_end()
    for x in instances[1:]:
        st = x['DBInstanceIdentifier']
        notice(f'deleting {st}')
        rds.delete_db_instance(
            DBInstanceIdentifier=st,
            SkipFinalSnapshot=True,
            DeleteAutomatedBackups=True,
        )
        time.sleep(5)
        notice_end()
    notice('deleting cluster')
    rds.delete_db_cluster(
        DBClusterIdentifier=name,
        SkipFinalSnapshot=True,
    )
    notice_end()



from raft import task
from ...base.utils import print_table, notice, notice_end, dump_yaml
from ..base import AwsTask, yielder


@task(klass=AwsTask)
def vpc_to_yaml(
        ctx, logical_name=None, vpc_id=None,
        session=None, quiet=False, **kwargs):
    """
    generates the cloudformation yaml for a vpc
    """
    ec2 = session.resource('ec2')
    vpc = ec2.Vpc(vpc_id)
    properties = dict(
        CidrBlock=vpc.cidr_block,
        InstanceTenancy=vpc.instance_tenancy,
        Tags=vpc.tags,
    )
    properties = { k: v for k, v in properties.items() if v is not None }
    data = {
        logical_name: dict(
            Type='AWS::EC2::VPC',
            DeletionPolicy='Retain',
            Properties=properties,
        ),
    }
    if not quiet:
        dump_yaml(data)
    return data


@task(klass=AwsTask)
def import_vpc(
        ctx, logical_name, vpc_id, filename,
        skip_review=False, session=None, prefix=None,
        **kwargs):
    """
    creates a yaml template for a vpc and imports it into cloudformation

    if skip_review is set, we will immediately try to execute the change set
      in cloudformation
    """
    from .import_resource_to_cfn import import_resource_to_cfn
    data = vpc_to_yaml(
        ctx, logical_name, vpc_id, session=session, quiet=True)
    resources_to_import = [dict(
        ResourceType='AWS::EC2::VPC',
        LogicalResourceId=list(data.keys())[0],
        ResourceIdentifier=dict(VpcId=vpc_id),
    )]
    import_resource_to_cfn(
        data, resources_to_import,
        filename, None, skip_review, session, prefix)


@task(klass=AwsTask)
def vpcs(ctx, session=None, raw=False, **kwargs):
    from ..ec2 import get_tag
    notice('loading vpcs')
    ec2 = session.client('ec2')
    rg_vpcs = list(yielder(ec2, 'describe_vpcs'))
    notice_end()
    if raw:
        dump_yaml(rg_vpcs, quiet=False)
    else:
        header = [ 'id', 'name', 'cidr', 'cidr6', ]
        rows = []
        for x in rg_vpcs:
            ipv4 = x.get('CidrBlockAssociationSet') or []
            ipv4 = [ cidr['CidrBlock'] for cidr in ipv4 ]
            ipv4 = ipv4 or [ x.get('CidrBlock') ]
            row = [
                x['VpcId'], get_tag(x, 'name') or '', ipv4[0],
            ]
            ipv6 = x.get('Ipv6CidrBlockAssociationSet')
            row.append((ipv6 or [{}])[0].get('Ipv6CidrBlock', ''))
            rows.append(row)
            for cidr in ipv4[1:]:
                rows.append([ '', '', cidr, '' ])
        print_table(header, rows)




from raft import Context
from raft.tasks import Task


class Api:
    def __init__(self, client_id, client_secret):
        from onelogin.api_client import ApiClient as OneLoginClient
        self.client = OneLoginClient(client_id, client_secret)


class OneLoginTask(Task):
    def __call__(self, *args, **kwargs):
        ctx = args[0]
        api = kwargs.get('api')
        if api is None and isinstance(ctx, Context):
            try:
                client_id = ctx.client_id
                client_secret = ctx.client_secret
                api = Api(client_id, client_secret)
            except AttributeError:
                pass
        kwargs['api'] = api
        return super().__call__(*args, **kwargs)



from decimal import Decimal
from functools import wraps
from io import StringIO
from itertools import chain
import os
import sys
from typing import Any
import yaml
from termcolor import colored, cprint


def expand_template(local_path, context):
    from jinja2 import Environment, FileSystemLoader
    jenv = Environment(loader=FileSystemLoader(os.getcwd()), lstrip_blocks=True, trim_blocks=True)
    content = jenv.get_template(local_path).render(**context or {})
    return content


def create_owner_string(owner=None, group=None):
    if owner:
        if group:
            return f'{owner}:{group}'
        return f'{owner}:'
    if group:
        return f':{group}'
    return None


def flush_print(st, **kwargs):
    """
    prints a string to stdout and flushes the stdout buffer.
    This is useful where stdout buffering delays the output of certain messages
    :param st: the message to print
    :param kwargs: additional kwargs to pass to the `print` function
    """
    flush = kwargs.pop('flush', True)
    print(st, **kwargs, flush=flush)
    sys.stdout.flush()


g_which_dot = -1


def dot_leader(st, length=40):
    """
    creates a dot leader to suffix a message with so that we get
    nicely aligned output and an easy way to get from message to value
    :param st: the message we are suffixing with the dot leader
    :param length: the field length (i.e., dot leader + len(st) == legnth)
    :return: the suffixed string, including the message + dot leader
    """
    global g_which_dot
    dot_chars = [ '.' ]
    g_which_dot += 1
    g_which_dot %= len(dot_chars)
    dot_char = dot_chars[g_which_dot]
    while length <= len(st):
        length += 10
    dots = dot_char * (length - len(st))
    return f'{st}{dots}'


def notice(st, length=40, **kwargs):
    """
    provides a notification message to the user.  this function is used
    in conjunction with `notice_end` to notify the user that something is
    happening and to await the response.
    :param st: the message that will be noticed to theuser
    :param length: the expected maximum field length for dot leads
    :param kwargs: additional kwargs to provide to `termcolor.cprint`
    :return:
    """
    ending = kwargs.pop('end', None) or ''
    color = kwargs.pop('color', 'blue')
    attrs = kwargs.pop('attrs', None)
    if isinstance(attrs, str):
        attrs = [ attrs ]
    cprint(dot_leader(st, length), end=ending, color=color, attrs=attrs, **kwargs)


def notice_end(st=None, color=None, **kwargs):
    """
    ends the notice cycle.  this function is used
    in conjunction with `notice` to notify the user that something is
    happening and to await the response.
    :param st: the message that will be provided to the user
    :param color: the color of the notification
    :param kwargs: additional kwargs to provide to `termcolor.cprint`
    :return:
    """
    if isinstance(st, str):
        cprint(st, color, **kwargs)
    elif st is False:
        flush_print('\N{wilted flower}')
    else:
        flush_print('\N{cherry blossom}')


def to_string(value: Any) -> str:
    """
    used by the `print_table` function, this function converts
    a value to a string making sure to format decimal.Decimal and float values
    to 2 decimal places
    :param Any value: the message that will be provided to the user
    :return: the string representation of value
    :rtype: str
    """
    if isinstance(value, (float, Decimal)):
        return f'{value:,.2f}'
    return f'{value}'


def print_table(header, rows, color='cyan'):
    """
    prints a header and rows in a tabular format in the specified color.
    this function is useful for presenting summary information from aws
    commands in a tabular format
    :param list[str] header: a list of strings for the table header
    :param list[Iterable[Any]] rows: a list of list of values to present
    :param str color: the color in which we will print our table
    """
    max_lens = [ 0 ] * len(header)
    rows = [ [ to_string(x) for x in row ] for row in rows ]
    for row in chain([ header ], rows):
        for i, x in enumerate(row):
            max_lens[i] = max(max_lens[i], len(x))
    f_header = []
    for width, x in zip(max_lens, header):
        diff = len(x)
        x = colored(x, color)
        diff = len(x) - diff
        f = f'%-{width + diff}s'
        f = f % (x,)
        f_header.append(f)
    print(' | '.join(f_header))
    print(' | '.join([ colored('-' * width, color) for width in max_lens ]))
    for row in rows:
        f_row = []
        for n, (width, x) in enumerate(zip(max_lens, row)):
            diff = 0
            if n == 0:
                diff = len(x)
                x = colored(x, color)
                diff = len(x) - diff
            f = f'%-{width + diff}s'
            f = f % (x,)
            f_row.append(f)
        print(' | '.join(f_row))


def confirm(message: str) -> bool:
    """
    presents the user with a confirmation message to which the user should
    respond with a YES / NO response
    :param message: the confirmation prompt
    :rtype: bool
    :return: whether the user entered yes or not
    """
    cprint(message, color='red', attrs=['underline'])
    confirmation = input("Please enter 'YES' to confirm and continue: ")
    return confirmation.lower() == 'yes'


def noticed(initial, completion=None):
    def wrap(fn):
        @wraps
        def decorated(*args, **kwargs):
            notice(initial)
            fn(*args, **kwargs)
            notice_end(completion)
        return decorated
    return wrap


def load_yaml(filename: str) -> Any:
    """
    Loads the yaml from filename and returns the relevant object
    :param filename: the name of the file to load
    :return: the constructed yaml object
    """
    with open(filename, 'r') as f:
        return yaml.load(f, yaml.SafeLoader)


def yaml_serializer():
    from ruamel.yaml import YAML
    y = YAML()
    y.indent(sequence=4, mapping=2, offset=2)
    return y


def get_context_value(ctx, key, default=None):
    pieces = key.split('.')
    value = ctx
    for piece in pieces:
        try:
            value = getattr(value, piece, None)
        except AttributeError:
            return default
    return value


def get_encrypted_context_value(ctx, session, key, default=None):
    from waddle import ParamBunch
    pieces = key.split('.')
    value = ctx
    for piece in pieces:
        try:
            value = getattr(value, piece, None)
        except AttributeError:
            return default
    value = ParamBunch.try_decrypt(value, session=session)
    return value


def dump_yaml(p, buff=None, quiet=True):
    """
    dumps an object as yaml to stdout
    """
    y = yaml_serializer()
    buff = buff or StringIO()
    y.dump(p, buff)
    buff.seek(0)
    if not quiet:
        print(buff.getvalue())
    return buff.getvalue()



import json
import os
from collections import defaultdict
from decimal import Decimal, ROUND_HALF_UP

from raft import task
from ...base.utils import notice, notice_end, print_table
from ..base import AwsTask, yielder
from ..base import name_tag
from ..ec2 import get_tag


def signature(data, region):
    instance_type = data['InstanceType']
    scope = data['Scope']
    if scope == 'Availability Zone':
        scope = data['AvailabilityZone']
    else:
        scope = region
    product = data['ProductDescription']
    if 'windows' in product.lower():
        product = 'windows'
    else:
        product = 'linux'
    return scope, product, instance_type


def instance_sig(instance):
    availability_zone = instance['Placement']['AvailabilityZone']
    product = instance.get('Platform', 'linux')
    product = product.lower()
    instance_type = instance['InstanceType']
    return availability_zone, product, instance_type


def described_instances(result):
    for r in result['Reservations']:
        for instance in r['Instances']:
            if instance['State']['Name'].lower() == 'running':
                yield instance


@task(klass=AwsTask)
def reservation_report(ctx, pricing=False, purchase=False, session=None, **kwargs):
    """
    produces an excel report of all instances and reservations

    The report will have two tabs:
      raw_data will show all of the instances, and the pricing comparison
         between reserved and on-demand pricing.  We use instance tags
         to determine which instances should be reserved.
         We will also try to do our best to indicate if we see a reservation
         in place that matches the instance so that we know which instances
         tagged reserved have a reservation in place and those that need one
         to purchase.
      the counts sheet will show all of the counts for reservations that we
         should purchase.  i.e., the breakdown by availability-zone and
         instance type of instances that are tagged reserved, but for
         which we do not see an active reservation.

    You can use the `--purchase` flag to initiate purchases of reservations
    in accordance with the counts sheet in the report.
    """
    from pandas import DataFrame
    from pandas import ExcelWriter
    client = session.client('ec2')
    notice('loading reservations')
    reservations = client.describe_reserved_instances()
    notice_end()
    notice('loading instances')
    result = client.describe_instances()
    instances = list(described_instances(result))
    while result.get('NextToken'):
        result = client.describe_instances(NextToken=result['NextToken'])
        instances += list(described_instances(result))
    notice_end(f'{len(instances)}')
    costs = defaultdict(lambda: defaultdict(dict))
    on_demand = defaultdict(lambda: defaultdict(dict))
    if pricing:
        costs, on_demand = reservation_pricing(ctx, session=session, quiet=True)
    sigs = []
    reservation_counts = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: 0)))
    reserved = []
    for reservation in reservations['ReservedInstances']:
        sig = signature(reservation, session.region_name)
        sigs.append(sig)
        if reservation['State'] in ('active', 'payment-pending'):
            n = reservation['InstanceCount']
            reservation_counts[sig[0]][sig[1]][sig[2]] += n

    region = session.region_name
    mantissa = Decimal('0.001')
    for instance in instances:
        sig = instance_sig(instance)
        n = reservation_counts[sig[0]][sig[1]][sig[2]]
        name = name_tag(instance)
        reservation = get_tag(instance, 'reserved').lower() == 'true'
        rg = costs[sig[0]][sig[1]][sig[2]]
        offering_id, hourly = ' ', Decimal(0)
        if rg and len(rg) == 2:
            offering_id, hourly = rg
        hourly = Decimal(hourly).quantize(mantissa, ROUND_HALF_UP)
        monthly = hourly * Decimal(730)
        on_demand_hourly = on_demand[sig[0]][sig[1]][sig[2]]
        on_demand_monthly = on_demand_hourly * Decimal(730)
        if n > 0 and reservation:
            reserved.append({
                'name': name,
                'status': 'reserved',
                # 'reservation_id': reservation_id,
                'instance_type': sig[2],
                'product': sig[1],
                'availability_zone': sig[0],
                'region': region,
                'reservation_tag': reservation,
                'offering_id': offering_id,
                'on_demand_hourly': float(on_demand_hourly),
                'on_demand_monthly': float(on_demand_monthly),
                'reserved_hourly': float(hourly),
                'reserved_monthly': float(monthly),
                'monthly': float(monthly if reservation else on_demand_monthly),
            })
            reservation_counts[sig[0]][sig[1]][sig[2]] -= 1
            continue
        n = reservation_counts[region][sig[1]][sig[2]]
        if n > 0 and reservation:
            reserved.append({
                'name': name,
                'status': 'reserved',
                # 'reservation_id': reservation_id,
                'instance_type': sig[2],
                'product': sig[1],
                'availability_zone': sig[0],
                'region': region,
                'reservation_tag': reservation,
                'offering_id': offering_id,
                'on_demand_hourly': float(on_demand_hourly),
                'on_demand_monthly': float(on_demand_monthly),
                'reserved_hourly': float(hourly),
                'reserved_monthly': float(monthly),
                'monthly': float(monthly if reservation else on_demand_monthly),
            })
            reservation_counts[region][sig[1]][sig[2]] -= 1
            continue
        reserved.append({
            'name': name,
            'status': 'not reserved',
            # 'reservation_id': reservation_id,
            'instance_type': sig[2],
            'product': sig[1],
            'availability_zone': sig[0],
            'region': region,
            'reservation_tag': reservation,
            'offering_id': offering_id,
            'on_demand_hourly': float(on_demand_hourly),
            'on_demand_monthly': float(on_demand_monthly),
            'reserved_hourly': float(hourly),
            'reserved_monthly': float(monthly),
            'monthly': float(monthly if reservation else on_demand_monthly),
        })
    for sig in sigs:
        n = reservation_counts[sig[0]][sig[1]][sig[2]]
        for x in range(n):
            reserved.append({
                'name': 'reservation',
                'status': 'unused',
                'instance_type': sig[2],
                'product': sig[1],
                'availability_zone': sig[0],
                'region': region,
                'reservation_tag': None,
                'offering_id': None,
                'on_demand_hourly': None,
                'on_demand_monthly': None,
                'reserved_hourly': None,
                'reserved_monthly': None,
                'monthly': None,
            })

    os.makedirs('./reports', exist_ok=True)
    profile = kwargs.get('prefix') or session.alias or kwargs.get('profile') or session.profile_name
    profile = profile.lower().replace(' ', '_').replace('-', '_')
    df = DataFrame(reserved)
    filename = f'./reports/{profile}_reservation_report.xlsx'
    writer = ExcelWriter(filename)  # pylint: disable=abstract-class-instantiated
    df.to_excel(writer, sheet_name='raw_data')
    df_counts = df[df['reservation_tag'] == True]  # pylint: disable=singleton-comparison
    df_counts = df_counts[df_counts['status'] == 'not reserved']
    header = [ 'availability_zone', 'instance_type', 'product', 'offering_id', ]
    df_counts = df_counts.groupby(header)
    df_counts = df_counts.size().reset_index()
    df_counts['n'] = df_counts[0]
    df_counts = df_counts[header + [ 'n' ]]
    df_counts.to_excel(writer, sheet_name='counts')
    writer.save()
    if not purchase:
        return
    to_purchase = df_counts.to_dict('records')
    for x in to_purchase:
        instance_type = x['instance_type']
        offering_id = x['offering_id']
        n = x['n']
        notice(f'purchasing {offering_id} / {instance_type} / {n}')
        response = client.purchase_reserved_instances_offering(
            ReservedInstancesOfferingId=offering_id,
            InstanceCount=n,
        )
        notice_end(response['ReservedInstancesId'])


@task(klass=AwsTask)
def on_demand_pricing(ctx, instances, session=None, **kwargs):
    """
    gets on-demand pricing for our instances
    """
    pricing = session.client('pricing', region_name='us-east-1')
    region = session.region_name
    costs = defaultdict(lambda: defaultdict(dict))
    for instance in instances:
        sig = instance_sig(instance)
        costs[sig[0]][sig[1]][sig[2]] = Decimal(0)
    prices = []
    for zone, v1 in costs.items():
        v1_values = list(v1.items())
        for product, v2 in v1_values:
            v2_values = list(v2.items())
            for instance_type, values in v2_values:
                filters = [
                    dict(Type='TERM_MATCH', Field='tenancy', Value='shared'),
                    dict(Type='TERM_MATCH', Field='operatingSystem', Value=product),
                    dict(Type='TERM_MATCH', Field='instanceType', Value=instance_type),
                    dict(Type='TERM_MATCH', Field='usagetype', Value=f'USE2-BoxUsage:{instance_type}'),
                    dict(Type='TERM_MATCH', Field='licenseModel', Value='No License required'),
                    dict(Type='TERM_MATCH', Field='preInstalledSw', Value='NA')
                ]
                if region not in ('us-east-1', 'us-west-2'):
                    filters.append(dict(Type='TERM_MATCH', Field='regionCode', Value=region))
                result = pricing.get_products(
                    ServiceCode='AmazonEC2',
                    Filters=filters, FormatVersion='aws_v1')
                if not result['PriceList']:
                    continue
                p = result['PriceList'][0]
                p = json.loads(p)
                for t1 in p['terms']['OnDemand'].values():
                    for t2 in t1['priceDimensions'].values():
                        hourly = t2['pricePerUnit']['USD']
                        hourly = Decimal(hourly).quantize(Decimal('0.0001'), ROUND_HALF_UP)
                        v2[instance_type] = hourly
                        prices.append({
                            'az': zone,
                            'product': product,
                            'type': instance_type,
                            'hourly': hourly,
                            'monthly': hourly * Decimal(365 * 2),
                            'yearly': hourly * Decimal(365 * 24),
                        })
    # header = [ 'az', 'product', 'type', 'hourly', 'monthly', 'yearly' ]
    # rows = []
    # for x in prices:
    #     row = [ x[key] for key in header ]
    #     rows.append(row)
    # print_table(header, rows)
    return costs


@task(klass=AwsTask)
def reservation_pricing(ctx, quiet=False, session=None, **kwargs):
    """
    compares up-front and reserved pricing costs
    """
    ec2 = session.client('ec2')
    costs = defaultdict(lambda: defaultdict(dict))
    on_demand_costs = None
    notice('instances')
    instances = list(yielder(ec2, 'describe_instances'))
    notice_end()
    for instance in instances:
        sig = instance_sig(instance)
        costs[sig[0]][sig[1]][sig[2]] = Decimal(0)
    i = 0
    offerings = []
    for zone, v1 in costs.items():
        v1_values = list(v1.items())
        for product, v2 in v1_values:
            v2_values = list(v2.items())
            for instance_type, values in v2_values:
                i += 1
                if not quiet:
                    notice(f'reservation info page {i}')
                d = 'Windows' if product == 'windows' else 'Linux/UNIX'
                result = ec2.describe_reserved_instances_offerings(
                    OfferingClass='standard',
                    InstanceTenancy='default',
                    InstanceType=instance_type,
                    MinDuration=365 * 24 * 60 * 60,
                    MaxDuration=365 * 24 * 60 * 60,
                    OfferingType='No Upfront',
                    ProductDescription=d,
                )
                for x in result['ReservedInstancesOfferings']:
                    if x['Scope'].lower() != 'region':
                        continue
                    amount = Decimal(x['RecurringCharges'][0]['Amount'])
                    amount = amount.quantize(Decimal('0.001'), ROUND_HALF_UP)
                    yearly = amount * Decimal(24 * 365)
                    monthly = Decimal(yearly) / Decimal(12)
                    hourly = Decimal(yearly) / Decimal(365 * 24)
                    st_id = x['ReservedInstancesOfferingId']
                    v2[instance_type] = (st_id, hourly)
                    offerings.append({
                        'az': zone,
                        'product': product,
                        'type': instance_type,
                        'hourly': hourly,
                        'monthly': monthly,
                        'yearly': yearly,
                    })
                if not quiet:
                    notice_end()
    if not quiet:
        header = [ 'az', 'product', 'type', 'hourly', 'monthly', 'yearly' ]
        rows = []
        for x in offerings:
            row = [ x[key] for key in header ]
            rows.append(row)
        print_table(header, rows)
    on_demand_costs = on_demand_pricing(ctx, instances, session=session)
    return costs, on_demand_costs



import os.path
from uuid import uuid4

from raft import task
from .base import XmlApiTask, XmlApi
from ..base.utils import notice, notice_end


def load_from_s3(session, url):
    notice(f'loading {url}')
    url = url[5:]
    bucket, key = url.split('/', 1)
    s3 = session.client('s3')
    response = s3.get_object(Bucket=bucket, Key=key)
    content = response['Body'].read()
    notice_end()
    return content


def load_file(filename):
    notice(f'loading {filename}')
    if os.path.exists(filename):
        with open(filename, 'rb') as f:
            content = f.read()
            notice_end()
            return content
    else:
        notice_end(False)
        return None


@task(klass=XmlApiTask)
def install_pfx(ctx, pfx, pfx_password=None,
                host=None, username=None, password=None,
                profile=None, **kwargs):
    """
    installs a cert to the firewall
    """
    from lxml import etree
    from OpenSSL.crypto import load_pkcs12
    api = XmlApi(host, username, password, profile=profile, **kwargs)
    session = kwargs['session']
    if pfx.startswith('s3://'):
        p12_data = load_from_s3(session, pfx)
    else:
        p12_data = load_file(pfx)
        if not p12_data:
            return

    p12 = load_pkcs12(p12_data, pfx_password)
    if not pfx_password:
        pfx_password = uuid4().hex.encode('utf-8')[:30]
        p12_data = p12.export(pfx_password)
    name = p12.get_certificate().get_subject()
    name = name.get_components()[0][-1].decode('utf-8')
    notice(f'importing to {host}')
    response = api.import_pfx(name, p12_data, pfx_password)
    success = response.attrib['status'] == 'success'
    notice_end(success)
    if not success:
        print(etree.tostring(response, pretty_print=True).decode())
    else:
        notice('committing')
        response = api.commit('updated ssl certificate')
        if response.tag == 'msg':
            line = response.xpath('.//line')
            if line:
                line = line[0]
                notice_end(line.text.lower())
            else:
                notice_end(False)
                print(etree.tostring(response, pretty_print=True).decode())
        else:
            notice_end(False)
            print(etree.tostring(response, pretty_print=True).decode())


@task(klass=XmlApiTask)
def install_pem(ctx, cert, key=None, passphrase=None,
                host=None, username=None, password=None,
                profile=None, **kwargs):
    """
    installs a cert / keypair to the firewall

    cert or keypair may be specified as an s3 url
    """
    from lxml import etree
    from OpenSSL.crypto import load_certificate
    from OpenSSL.crypto import load_privatekey
    from OpenSSL.crypto import dump_privatekey
    from OpenSSL.crypto import FILETYPE_PEM
    api = XmlApi(host, username, password, profile=profile, **kwargs)
    session = kwargs['session']
    notice(f'loading {cert}')
    if cert.startswith('s3://'):
        cert_content = load_from_s3(session, cert)
    else:
        cert_content = load_file(cert)
        if not cert_content:
            return

    key_content = None
    if key:
        if key.startswith('s3://'):
            key_content = load_from_s3(session, key)
        else:
            key_content = load_file(key)
            if not key_content:
                return
        pem_key = load_privatekey(FILETYPE_PEM, key_content, passphrase=passphrase)
        if not passphrase:
            passphrase = uuid4().hex.encode('utf-8')[:30]
            key_content = dump_privatekey(
                FILETYPE_PEM,
                pem_key,
                passphrase=passphrase)

    pem = load_certificate(FILETYPE_PEM, cert_content)
    name = pem.get_subject()
    name = name.get_components()[0][-1].decode('utf-8')
    notice(f'importing to {host}')
    response = api.import_pem(
        name,
        cert_content,
        key_content,
        passphrase=passphrase)
    success = False
    try:
        success = response.attrib['status'] == 'success'
        notice_end(success)
    except KeyError:
        notice_end(success)
    if not success:
        print(etree.tostring(response, pretty_print=True).decode())
    else:
        notice('committing')
        response = api.commit('updated ssl certificate')
        if response.tag == 'msg':
            line = response.xpath('.//line')
            if line:
                line = line[0]
                notice_end(line.text.lower())
            else:
                notice_end(False)
                print(etree.tostring(response, pretty_print=True).decode())
        else:
            notice_end(False)
            print(etree.tostring(response, pretty_print=True).decode())



from raft.tasks import Task
from raft.context import Context


class AzureTask(Task):
    def __call__(self, *args, **kwargs):
        from azure.identity import ClientSecretCredential
        from azure.identity import DefaultAzureCredential
        from ..base.utils import get_context_value
        from ..base.utils import notice, notice_end
        ctx = args[0]
        has_context = isinstance(ctx, Context)
        client_id = kwargs.get('client_id')
        client_secret = kwargs.get('client_secret')
        tenant_id = kwargs.get('tenant_id')
        creds = kwargs.get('creds')
        if has_context:
            client_id = client_id or get_context_value(ctx, 'azure.client_id')
            client_secret = client_secret or get_context_value(ctx, 'azure.client_secret')
            tenant_id = tenant_id or get_context_value(ctx, 'azure.tenant_id')
        notice('client_id')
        notice_end(client_id)
        notice('tenant_id')
        notice_end(tenant_id)
        if not creds and client_id and client_secret and tenant_id:
            creds = ClientSecretCredential(tenant_id, client_id, client_secret)
        else:
            notice('no credentials found')
            creds = DefaultAzureCredential()
            notice_end()
        kwargs['creds'] = creds
        return super().__call__(*args, **kwargs)





#!/usr/bin/env python

# Support setuptools only, distutils has a divergent and more annoying API and
# few folks will lack setuptools.
from setuptools import setup
from importlib.resources import open_text
# Version info -- read without importing
_locals = {}

version = open_text('{{ name }}', 'version.txt').read().strip()

# PyYAML ships a split Python 2/3 codebase. Unfortunately, some pip versions
# attempt to interpret both halves of PyYAML, yielding SyntaxErrors. Thus, we
# exclude whichever appears inappropriate for the installing interpreter.
exclude = ["*.yaml2", 'test']

# Frankenstein long_description: version-specific changelog note + README
with open('README.md') as f:
    long_description = f.read()

extras = {}
all_extras = set()
for x in [ 'dev' ]:
    filename = f'{x}.txt'
    with open(filename, 'r') as f:
        st = f.read()
    rg = st.split()
    extras[x] = rg
    if x != 'dev':
        all_extras |= set(rg)
if all_extras:
    all_extras = list(all_extras)
    all_extras.sort()
    extras['all'] = all_extras


setup(
    name='{{ name }}',
    version=version,
    description='{{ name }}',
    license='BSD',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='dev',
    author_email='developers@directbuy.com',
    url='https://bitbucket.org/dbuy/{{ name }}',
    packages=[
        '{{ name }}',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=[
    ],
    extras_require=extras,
)



from raft.tasks import task
from raft.collection import Collection
from . import build as build_tasks


@task
def setup(ctx):
    """
    installs dependencies in the codebuild environment
    this installation is dead simple because we don't have
    to worry about python virtual environments
    """
    pip = 'pip install -q --user'
    ctx.run(f'{pip} -U wheel')
    ctx.run(f'{pip} -U setuptools raft pip')
    ctx.run(f'{pip} -r requirements.txt')


codebuild_collection = Collection('codebuild')
codebuild_collection.add_task(setup)
global_collection = Collection()
global_collection.add_tasks(build_tasks)



import json
import os
import time
from collections.abc import Mapping
from configparser import ConfigParser
from raft import task
from raft.config import DataProxy

from ..aws.base import AwsTask
from ..base.utils import notice_end, notice, get_context_value, \
    get_encrypted_context_value, dump_yaml


def airflow_config(ctx, filename):
    if filename:
        notice(f'checking {filename}')
        if not os.path.exists(filename):
            notice_end(False)
            filename = None
        else:
            notice_end()
    if not filename:
        notice('checking config for airflow.cfg')
        filename = get_context_value(ctx, 'airflow.config')
        if filename:
            if not os.path.exists(filename):
                filename = None
                notice_end('warning, convocations config file incorrect')
            else:
                notice_end()
        else:
            notice_end(False)
        if not filename:
            notice('checking AIRFLOW_HOME')
            t = os.environ.get('AIRFLOW_HOME')
            if t:
                t = os.path.join(t, 'airflow.cfg')
                if os.path.exists(t):
                    filename = t
            notice_end(bool(filename))
    return filename


@task
def generate_fernet_key(ctx, filename=None):
    """
    generates a fernet key for the airflow.cfg file
    if no fernet key exists within the current configuration.  If a fernet
    key exists within the current configuration, we simply ignore.
    """
    filename = airflow_config(ctx, filename)
    p = ConfigParser()
    p.read(filename)
    fernet_key = p['core'].get('fernet_key')
    if not fernet_key:
        from cryptography.fernet import Fernet
        fernet_key = Fernet.generate_key().decode()
        p['core']['fernet_key'] = fernet_key
        with open(filename, 'w') as f:
            p.write(f)


def traverse_and_decrypt(d, session):  # pylint: disable=inconsistent-return-statements
    from waddle import ParamBunch
    if not isinstance(d, (Mapping, DataProxy)):
        value = ParamBunch.try_decrypt(d, session=session)
        return value
    for key, value in d.items():
        if isinstance(value, Mapping):
            new_value = dict(traverse_and_decrypt(value, session))
            yield key, new_value
        elif isinstance(value, list):
            new_value = []
            for x in value:
                x = ParamBunch.try_decrypt(x, session=session)
                new_value.append(x)
            yield key, new_value
        else:
            new_value = ParamBunch.try_decrypt(value, session=session)
            yield key, new_value
    return


@task(klass=AwsTask)
def setup_local_connections(ctx, key=None, session=None, debug=False, **kwargs):
    """
    installs connections on your local airflow
    """
    import requests
    password = get_encrypted_context_value(ctx, session, 'airflow.password')
    conf = ctx.airflow
    aws_session = session
    session = requests.Session()
    session.auth = (conf.username, password)
    if debug:
        notice(f'connecting to api as {conf.username} / {password}')
    else:
        notice(f'connecting to api as {conf.username}')
    base_url = f'{conf.base_url}/api/v1/connections'
    response = session.get(base_url)
    notice_end(response.status_code == 200)
    if response.status_code != 200:
        notice_end(response.text)
        return
    data = response.json()
    connections = conf.connections
    existing_connections = data['connections']
    existing_connection_ids = {
        x['connection_id'] for x in existing_connections
    }
    for connection_id in connections.keys():
        if key and key != connection_id:
            continue
        x = connections[connection_id]
        found = connection_id in existing_connection_ids
        rg = list(traverse_and_decrypt(x, aws_session))
        x = dict(rg)
        if found:
            fn = session.patch
            url = f'{base_url}/{connection_id}'
            notice(f'patching {connection_id}')
        else:
            fn = session.post
            url = base_url
            notice(f'creating {connection_id}')
        body = dict(
            connection_id=connection_id,
            conn_type=x.get('conn_type') or '',
            host=x.get('host') or '',
            login=x.get('login') or '',
            password=x.get('password') or '',
            schema=x.get('schema') or '',
            port=x.get('port') or None,
            description=x.get('description') or None,
        )
        extra = x.get('extra')
        if extra:
            extra = json.dumps(extra)
            body['extra'] = extra
        response = fn(url, json=body)
        notice_end(response.status_code == 200)
        if response.status_code != 200:
            dump_yaml(response.json(), quiet=False)


@task(klass=AwsTask)
def setup_local_variables(ctx, session=None, **kwargs):
    """
    installs variables on your local airflow running in docker compose
    """
    from murmuration import kms_wrapped
    import requests
    notice('getting variables from configuration')
    variables = get_context_value(ctx, 'airflow.variables')
    if not variables:
        notice_end(False)
        return
    notice_end()
    conf = ctx.airflow
    password = get_encrypted_context_value(ctx, session, 'airflow.password')
    api_session = requests.Session()
    api_session.auth = (conf.username, password)
    base_url = f'{conf.base_url}/api/v1/variables'
    response = api_session.get(base_url)
    data = response.json()
    existing_variables = { x['key'] for x in data['variables'] }
    for key, value in variables.items():
        found = key in existing_variables
        if found:
            fn = api_session.patch
            url = f'{base_url}/{key}'
            notice(f'patching {key}')
        else:
            fn = api_session.post
            url = base_url
            notice(f'creating {key}')
        body = dict(key=key, value=value)
        response = fn(url, json=body)
        notice_end(response.status_code == 200)
        if response.status_code != 200:
            dump_yaml(response.json(), quiet=False)


@task
def clear_queue(ctx, filename=None):
    """
    clears the celery queue to prevent weird errors
    `TypeError: unsupported operand type(s) for +: 'float' and 'str'`
    """
    from redis import Redis
    filename = filename or ctx.airflow.config
    p = ConfigParser()
    p.read(filename)
    notice('looking up broker url')
    broker_url = p['celery'].get('broker_url')
    if not broker_url:
        cmd = p['celery'].get('broker_url_cmd')
        if cmd:
            result = ctx.run(cmd)
            broker_url = result.stdout.strip()
    if not broker_url:
        notice_end('not found')
        return
    notice_end()
    notice('clearing default queue')
    client = Redis(broker_url)
    client.delete('default')
    notice_end()


@task
def drain_celery(ctx, config=None, broker=None, timeout=900, poll=5):
    """
    drains celery by canceling consumers and waits for all active tasks
    to complete.
    """
    config = airflow_config(ctx, config)
    if not broker:
        notice(f'reading {config}')
        p = ConfigParser()
        p.read(config)
        notice_end()
        broker = p['celery']['broker_url']
    from celery import Celery  # pylint: disable=import-error
    app = Celery(broker=broker)
    notice('finding active queues')
    inspector = app.control.inspect()
    result = inspector.active_queues() or {}
    if not result:
        notice_end('no queues found')
        return
    active_queues = set()
    for queues in result.values():
        for queue in queues:
            active_queues.add(queue['name'])
    active_queues = list(active_queues)
    notice_end(len(active_queues))
    for queue in active_queues:
        notice(f'canceling {queue}')
        app.control.cancel_consumer(queue)
        notice_end()
    for x in range(timeout // poll):
        notice(f'active tasks [{x + 1}/{timeout // poll}]')
        active = inspector.active()
        n = 0
        for tasks in active.values():
            n += len(tasks)
        notice_end(n)
        if n == 0:
            break
        time.sleep(poll)



from raft import task

from ..cfn.import_instance import get_instance
from ..base import AwsTask


@task(klass=AwsTask)
def used_ips(ctx, subnet=None, verbose=False, session=None, **kwargs):
    """
    shows all ips used in a particular subnet in a tabular format

    * you can provide either the subnet id or a substring of the subnet's
      name tag
    * if you provide the name, and multiple subnets match that same name,
      all ips from all matching subnets will be shown; to avoid this,
      use the subnet id in the `subnet` parameter
    * matches will be shown in ascending order of ip address
    * will show the name and id of ec2 instances
    * will show the elb name and elb type of elastic load balancers
    * will show the file system id for efs mount targets
    * if verbose mode is enabled, if a resource is not an efs, elb, or ec2
      instance, we will print the full yaml for the object
    """
    from ...base.utils import notice, notice_end
    from ...base.utils import dump_yaml
    from ...base.utils import print_table
    from ..ec2 import get_tag
    from ..base import yielder
    ec2 = session.client('ec2')
    notice('loading subnets')
    rg_subnets = list(yielder(ec2, 'describe_subnets', session=session))
    notice_end()
    id_matches = []
    name_matches = []
    matches = None
    subnet = (subnet or '').lower()
    for x in rg_subnets:
        n_id = x['SubnetId']
        if n_id == subnet:
            id_matches.append(n_id)
        name = get_tag(x, 'name')
        if subnet in name.lower():
            name_matches.append(n_id)
    if id_matches:
        matches = id_matches
    else:
        matches = name_matches
    if not matches:
        notice_end('no matching subnets')
    else:
        notice_end(f'found {len(matches)} subnet match(es)')
    notice('loading interfaces')
    rg_interfaces = list(yielder(
        ec2,
        'describe_network_interfaces',
        session=session,
        Filters=[{
            'Name': 'subnet-id',
            'Values': matches,
        }]
    ))
    notice_end(f'found {len(rg_interfaces)}')
    notice('verbose')
    notice_end(f'{verbose}')
    header = [ 'ip', 'id', 'name' ]
    rows = []
    for x in rg_interfaces:
        attachment = x.get('Attachment') or {}
        instance_id = attachment.get('InstanceId')
        owner_id = attachment.get('InstanceOwnerId')
        description = (x.get('Description') or '').lower()
        interface_type = x.get('InterfaceType') or ''
        name = ''
        if instance_id:
            name, _ = get_instance(ctx, None, instance_id, session=session)
        elif owner_id == 'amazon-elb':
            instance_id, name = description.rsplit('/', 1)[0].rsplit('/', 1)
        elif owner_id == 'amazon-aws':
            if description.startswith('efs'):
                description = description.rsplit(' ', 1)[0]
                name, instance_id = description.split(' for ', 1)
            if interface_type == 'vpc_endpoint':
                instance_id = description.split(' ')[-1]
                name = interface_type
        if verbose and not instance_id:
            dump_yaml(x, quiet=False)
        if instance_id is None:
            instance_id = x['NetworkInterfaceId']
            name = interface_type
        for y in x['PrivateIpAddresses']:
            rows.append([
                y['PrivateIpAddress'],
                instance_id,
                name,
            ])
    rows.sort(key=lambda lx: list(map(int, lx[0].split('.'))))
    print_table(header, rows)



####
#  To help us create environment-specific keypairs
#  that automatically get added to ssm's parameter store
#  for use in deploys
####
import boto3
from raft.tasks import task
from .base import AwsTask
from ..base.utils import notice_end, notice


__all__ = [
    'new_key_pair',
    'show_public_key',
]


@task(klass=AwsTask, help=dict(
    name='the name of the app',
    namespace='the ssm and keypair namespace',
    session='leave this blank, it will be filled in by the context',
))
def new_key_pair(
        ctx, name, namespace, key_type='ed25519',
        param_type='SecureString', key_arn=None,
        session=None, **kwargs):
    """
    creates a new key pair and stores it in aws ssm
    the ssm key for the private key will be /namespace/name
    while the name of the keypair will be  namespace/name

    This function to allows us to create environment-
      specific deploy keys that would allow us to, for example, do
      deploys from codebuild a little more easily

    Args:

        ctx: ignore, used internally by convoke
        name (str):
            the name of the key pair
        namespace (str):
            the namespace of the keypair, the namespace of the ssm
            parameter, and the alias to kms key
        key_type (str):
            specifies the type of key to create (rsa or ed25519),
            defaults to ed25519
        param_type (str):
            specifies the type of SSM parameter (String or SecureString)
        key_arn (str):
            if you want to override looking for a key with the
            alias `namespace`, then specify the key arn explicitly.
        session: ignore, used internally
    """
    session = session or boto3.Session(**kwargs)
    ec2 = session.client('ec2')
    if not name.endswith('deploy_key'):
        name = f'{name}_deploy_key'
    key_name = f'{namespace}/{name}'
    notice('creating key pair')
    response = ec2.create_key_pair(KeyName=key_name, KeyType=key_type)
    private_key = response['KeyMaterial']
    parameter_name = f'/{key_name}'
    notice_end(key_name)

    kms = session.client('kms')
    key_id = None
    if key_arn:
        notice(f'key {key_arn}?')
        try:
            response = kms.describe_key(KeyId=key_arn)
            notice_end()
            key_id = response['KeyMetadata']['KeyId']
        except:
            notice_end('not found')

    if not key_id:
        notice(f'key aliased {namespace}?')
        try:
            response = kms.describe_key(KeyId=f'alias/{namespace}')
            key_id = response['KeyMetadata']['KeyId']
            notice_end()
        except:
            notice_end('not found')

    notice('creating ssm parameter')
    ssm = session.client('ssm')
    params = dict(
        Name=parameter_name,
        Description=f'[{namespace}] deployment key for {name}',
        Value=private_key,
        Type=param_type,
    )
    if param_type == 'SecureString' and key_id:
        params['KeyId'] = key_id
    ssm.put_parameter(**params)
    notice_end()


@task(klass=AwsTask, help=dict(
    name='the name of the app',
    namespace='the ssm and keypair namespace',
    session='leave this blank, it will be filled in by the context',
))
def show_public_key(ctx, name, namespace, session=None, **kwargs):
    """
    shows the public key for the /namespace/name private key
    generated from `new_key_pair`

    Args:

        ctx: ignore, used internally by convoke
        name (str):
            the name of the key pair
        namespace (str):
            the namespace of the keypair, the namespace of the ssm
            parameter, and the alias to kms key
        session: ignore, used internally
    """
    from cryptography.hazmat.primitives import serialization as crypto_serialization

    session = session or boto3.Session(**kwargs)
    if not name.endswith('deploy_key'):
        name = f'{name}_deploy_key'
    key_name = f'{namespace}/{name}'
    parameter_name = f'/{key_name}'
    notice('getting key from ssm')
    ssm = session.client('ssm')
    response = ssm.get_parameter(Name=parameter_name, WithDecryption=True)
    value = response['Parameter']['Value']
    notice_end()
    private_key = crypto_serialization.load_ssh_private_key(
        value.encode('utf-8'),
        password=None,
    )
    public_key = private_key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH,
        crypto_serialization.PublicFormat.OpenSSH
    )
    print(public_key.decode('utf-8'))




import os
from glob import iglob

from raft import task


@task
def upload_files(ctx, source, site_name, destination):
    """
    uploads files specified by source to a specified sharepoint site folder

    source can be a glob
    """
    import adal
    from office365.runtime.auth.token_response import TokenResponse
    from office365.sharepoint.client_context import ClientContext
    from office365.sharepoint.folders.folder import Folder
    sharepoint_base_url = f'https://{ctx.tenant_name}.sharepoint.com'
    site_url = f'{sharepoint_base_url}/sites/{site_name}'
    authority = f'https://login.microsoftonline.com/{ctx.tenant_id}'
    app = adal.AuthenticationContext(authority)
    token_json = app.acquire_token_with_client_certificate(
        sharepoint_base_url,
        ctx.client_id,
        ctx.key,
        ctx.thumbprint,
    )
    session = ClientContext(site_url).with_access_token(
        lambda: TokenResponse.from_json(token_json)
    )
    destination = destination.replace('Documents', 'Shared Documents')
    destination = os.path.join('/sites', site_name, destination)
    folder = session.web.get_folder_by_server_relative_path(destination)
    folder: Folder = folder.get().execute_query()
    for x in iglob(source):
        print(f'uploading {x} => {folder.name}')
        with open(x, 'rb') as f:
            folder.upload_file(os.path.basename(x), f.read()).execute_query()



from raft.tasks import task
from .base import AzureTask


@task(klass=AzureTask)
def resource_groups(
        ctx, name=None,
        client_id=None, client_secret=None,
        subscription_id=None, creds=None, **kwargs):
    """
    prints a list of available resource groups in a subscription
    """
    from convocations.base.utils import get_context_value
    from convocations.base.utils import print_table
    from azure.mgmt.resource.resources import ResourceManagementClient
    name = name or ''
    subscription_id = subscription_id or get_context_value(ctx, 'azure.subscription_id')
    client = ResourceManagementClient(credential=creds, subscription_id=subscription_id)
    header = [ 'name' ]
    rows = []
    for x in client.resource_groups.list():
        st = x.name.lower()
        if name in st:
            rows.append([
                x.name
            ])
    print_table(header, rows)


@task(klass=AzureTask)
def providers(
        ctx, name=None, verbose=False,
        client_id=None, client_secret=None,
        subscription_id=None, creds=None, **kwargs):
    """
    prints a list of available providers and resource types
    """
    from convocations.base.utils import get_context_value
    from convocations.base.utils import print_table, notice, notice_end
    from azure.mgmt.resource.resources import ResourceManagementClient
    name = name or ''
    subscription_id = subscription_id or get_context_value(ctx, 'azure.subscription_id')
    notice('loading providers')
    client = ResourceManagementClient(credential=creds, subscription_id=subscription_id)
    if verbose:
        header = [ 'provider', 'resource_type', 'api_version' ]
    else:
        header = [ 'provider', 'resource_type', 'default_api_version' ]
    rows = []
    name = name.lower()
    for x in client.providers.list():
        ns = x.namespace
        st = ns.lower()
        if name in st:
            if verbose:
                for y in x.resource_types:
                    for v in y.api_versions:
                        rows.append([ ns, y.resource_type, v ])
            else:
                for y in x.resource_types:
                    rows.append([
                        ns,
                        y.resource_type,
                        y.default_api_version,
                    ])
    notice_end()
    print_table(header, rows)



import time
from raft import task
from convocations.palo_alto.base import XmlApiTask


@task(klass=XmlApiTask)
def upgrade(ctx, version='10.1.5', api=None, username=None, password=None,
            host=None, peer=None, session=None, **kwargs):
    """
    Updates the ha pair
    """
    from panos.firewall import Firewall
    from .base import XmlApi
    from ..base.utils import notice, notice_end
    from lxml import etree
    api = api or XmlApi(host, username, password, session=session)
    username, password = XmlApi.resolve_creds(username, password, session)
    primary = Firewall(host, username, password)
    secondary = None
    if secondary:
        secondary = Firewall(peer, username, password)
        notice('setting preemptive to no')
        params = dict(
            xpath=(
                '/config/devices/entry[@name="localhost.localdomain"]'
                '/deviceconfig/high-availability/group/election-option'
            ),
            element='<preemptive>no</preemptive>',
        )
        r = api.post('config', 'set', **params)
        print(etree.tostring(r, pretty_print=True).decode('utf-8'))
        primary.commit(sync=True)
        notice_end()
        notice(f'upgrading secondary to version {version}')
        secondary.software.upgrade_to_version(version)
        notice_end()
        notice('waiting 120 seconds for secondary to come back up')
        time.sleep(120)
        notice_end()
        notice('checking version after upgrade')
        secondary.refresh_system_info()
        notice_end(secondary.version)
        input('please press <enter> to continue')
        notice('suspending ha on active peer')
        primary.op('request high-availability state suspend')
        notice_end()

    notice(f'upgrading primary to version {version}')
    primary.software.upgrade_to_version(version)
    notice_end()
    notice('waiting 120 seconds for primary to come back up')
    time.sleep(120)
    notice_end()
    notice('checking version after upgrade')
    primary.refresh_system_info()
    notice_end(primary.version)

    if secondary:
        notice('setting preemptive to yes')
        params = dict(
            xpath=(
                '/config/devices/entry[@name="localhost.localdomain"]'
                '/deviceconfig/high-availability/group/election-option'
            ),
            element='<preemptive>yes</preemptive>',
        )
        r = api.post('config', 'set', **params)
        notice_end()
        notice('committing')
        primary.commit(True)
        notice_end()
        print(etree.tostring(r, pretty_print=True).decode('utf-8'))
        notice('turning on ha on primary')
        primary.op('request high-availability state functional')
        notice_end()
        notice('turning off ha on secondary to force failover')
        secondary.op('request high-availability state suspend')
        notice_end()
        time.sleep(10)
        notice('turning ha back on secondary')
        secondary.op('request high-availability state functional')
        notice_end()



import os
import os.path
from tempfile import NamedTemporaryFile
from raft.tasks import task


@task
def ed25519_self_signed_cert(
        ctx, names, days=365, directory='.', outfile=None):
    """
    creates a new self-signed ed25519 cert with the appropriate names as subjects
      * names should be specified as a comma separated list, with no spaces
        between the commas
      * the first `name` is used as the subject
      * outputs key to outfile.key, cert to outfile.crt, and csr to outfile.csr
      * if no outfile is specified, will use the subject as the outfile base
    :param ctx:
    :param names: the subject / alt names to use with the
    :param str curve: pick one of the available values from
        `openssl ecparam -list_curves`
    :param int days: number of days the cert should last, default is 365
    :param str directory: the directory in which the cert, key, and csr
        will be stored
    :param outfile:
    """
    names = names.split(',')
    outfile = outfile or names[0]
    key_path = os.path.join(directory, f'{outfile}.key')
    cert_path = os.path.join(directory, f'{outfile}.crt')
    csr_path = os.path.join(directory, f'{outfile}.csr')
    ctx.run(f'openssl genpkey -algorithm ED25519 >{key_path}')
    with NamedTemporaryFile('w') as f:
        f.write('\n'.join([
            '[req]',
            'distinguished_name = req_distinguished_name',
            'req_extensions = v3_req',
            'prompt = no',
            '[req_distinguished_name]',
            'C = US',
            f'CN = {names[0]}',
            '[v3_req]',
            'keyUsage = keyEncipherment, dataEncipherment',
            'extendedKeyUsage = serverAuth, clientAuth',
            'subjectAltName = @alt_names',
            '[alt_names]',
            ''
        ]))
        for i, x in enumerate(names, 1):
            f.write(f'DNS.{i} = {x}\n')
        f.flush()
        command = (
            f'openssl req -new'
            f' -out {csr_path}'
            f' -key {key_path}'
            f' -config {f.name}'
        )
        ctx.run(command)
        command = (
            f'openssl x509 -req -days {days}'
            f' -in {csr_path} -signkey {key_path} -out {cert_path}'
        )
        ctx.run(command)



from raft import task
from ..base import AwsTask
from ..base import yielder
from ..base import name_tag


@task(klass=AwsTask)
def efs(ctx, name=None, access_points=True, session=None, **kwargs):
    """
    lists matching efs drives where the value of name tag contains the provided value
    if no value for `name` is provided, the convocation will display all
    efs drives in the region.
    """
    from ...base.utils import print_table, notice, notice_end
    client = session.client('efs')
    notice('describing file systems')
    fn = yielder(client, 'describe_file_systems', session=session)
    shares = list(fn)
    notice_end(len(shares))
    header = [ 'id', 'name', 'date', ]
    if access_points:
        header.append('access points')
    rows = []
    name = (name or '').lower()
    for x in shares:
        fs_name = name_tag(x)
        fs_id = x['FileSystemId']
        if name in fs_name or name == fs_id:
            row = [fs_id , fs_name, x['CreationTime'] ]
            rows.append(row)
    if access_points:
        notice('describing access points')
        for row in rows:
            fs_id = row[0]
            try:
                response = client.describe_access_points(FileSystemId=fs_id)
            except:
                row.append('')
                continue
            if not response['AccessPoints']:
                row.append('')
                continue
            ap_ids = [ x['AccessPointId'] for x in response['AccessPoints'] ]
            row.append(', '.join(ap_ids))
        notice_end()
    print_table(header, rows)



# pylint: disable=redefined-builtin
from raft.tasks import task

from ...base.utils import print_table
from ..base import AwsTask
from . import get_tag


@task(klass=AwsTask, help=dict(
    id='[optional] a comma delimited list of ids',
))
def route_tables(ctx, name=None, ids=None, session=None, **kwargs):
    """
    prints a table of all routes in the route tables matching the `ids` argument.
    route tables can be specified as a comma-delimited list.  if no ids
    are specified, we will display all routes in all route tables.

    as an alternative to an id, you may specify the name of the route table
    as identified by the `Name` tag.  we will include any route tables where
    the specified name is a substring of the tag value.
    """
    ec2 = session.client('ec2')
    if ids:
        ids = ids.split(',')
        rg = ec2.describe_route_tables(RouteTableIds=ids)
    else:
        rg = ec2.describe_route_tables()
    name = (name or '').lower()
    header = [ 'id', 'name', 'dest', 'next_hop', ]
    rows = []
    for x in rg['RouteTables']:
        route_table_name = (get_tag(x, 'Name') or '').lower()
        if name not in route_table_name:
            continue
        for route in x['Routes']:
            rows.append([
                x['RouteTableId'],
                route_table_name,
                route.get('DestinationCidrBlock') or route.get('DestinationIpv6CidrBlock') or '',
                route.get('GatewayId') or route.get('InstanceId') or
                route.get('NetworkInterfaceId') or route.get('TransitGatewayId') or
                route.get('VpcPeeringConnectionId') or ''
            ])
    print_table(header, rows)



from raft import task

from convocations.base.utils import notice_end, notice, print_table, dump_yaml
from ..base import AwsTask, yielder
from .import_resource_to_cfn import sync_resource


@task(klass=AwsTask)
def pipelines(ctx, name=None, **kwargs):
    """
    lists the code pipelines available in an account / region

    name: an optional substring, case-insensitive matcher (non-regex)
    """
    notice('getting pipelines')
    session = kwargs.pop('session')
    cp = session.client('codepipeline')
    response = list(yielder(cp, 'list_pipelines'))
    notice_end()
    rows = []
    name = (name or '').lower()
    for x in response:
        if name in x['name'].lower():
            rows.append([ x['name'], x['updated'].strftime('%Y-%m-%d') ])
    header = [ 'name', 'updated' ]
    print()
    print_table(header, rows)
    print()


@task(klass=AwsTask)
def code_build_projects(ctx, name=None, session=None, **kwargs):
    """
    lists the code projects available in an account / region

    name: an optional substring, case-insensitive matcher (non-regex)
    """
    notice('getting projects')
    cp = session.client('codebuild')
    response = list(yielder(cp, 'list_projects', session, sortBy='NAME'))
    notice_end()
    rows = []
    name = (name or '').lower()
    for x in response:
        if name in x.lower():
            rows.append([ x ])
    header = [ 'name' ]
    print()
    print_table(header, rows)
    print()


@task(klass=AwsTask)
def sync_pipeline(ctx, name, filename, logical_name=None, **kwargs):
    """
    prints the cloudformation for the current version of the pipeline
    named name so that we can remove drift from cloudformation

    :param str name: the name of the code pipeline in aws
    :param str filename: the yaml filename of the template without the `profile/cfn` prefix
    :param str logical_name: the logical name of the resource in the cloudformation template
    :param Context ctx: the convocation context
    """
    from .import_resource_to_cfn import to_pascal_case
    session = kwargs['session']
    cp = session.client('codepipeline')
    response = cp.get_pipeline(name=name)
    pline = response['pipeline']
    logical_name = logical_name or ''.join(map(lambda lx: lx.title(), name.split('-')))
    properties = dict(
        ArtifactStore=to_pascal_case(pline.get('artifactStore')) or None,
        ArtifactStores=to_pascal_case(pline.get('artifactStores')) or None,
        DisableInboundStageTransitions=None,
        Name=pline['name'],
        RestartExecutionOnUpdate=False,
        RoleArn=pline.get('roleArn'),
        Stages=to_pascal_case(pline.get('stages')),
        Tags=None,
    )
    properties = { k: v for k, v in properties.items() if v is not None }
    data = {
        logical_name: dict(
            Type='AWS::CodePipeline:Pipeline',
            DeletionPolicy='Retain',
            Properties=properties,
        ),
    }
    sync_resource(filename, logical_name, data)


@task(klass=AwsTask)
def code_build_project_yaml(ctx, name, quiet=False, **kwargs):
    """
    prints the cloudformation for the current version of the pipeline
    named name so that we can remove drift from cloudformation

    :param str name: the name of the code pipeline in aws
    :param bool quiet: set this flag to prevent output to stdout / stderr and return the yaml dictionary
    :param Context ctx: the convocation context
    """
    from .import_resource_to_cfn import to_pascal_case, filtered_tags
    session = kwargs['session']
    cb = session.client('codebuild')
    response = cb.batch_get_projects(names=[ name ])
    project = response['projects'][0]
    logical_name = ''.join(map(lambda lx: lx.title(), name.split('-')))
    tags = filtered_tags(project)
    properties = dict(
        Name=project['name'],
        Description=project.get('description'),
        Source=to_pascal_case(project['source']),
        SecondarySources=to_pascal_case(project.get('secondarySources')) or None,
        SourceVersion=project.get('sourceVersion'),
        SecondarySourceVersions=to_pascal_case(project.get('secondarySourceVersions')) or None,
        Artifacts=to_pascal_case(project.get('artifacts')) or None,
        SecondaryArtifacts=to_pascal_case(project.get('secondaryArtifacts')) or None,
        Cache=to_pascal_case(project.get('cache')) or None,
        Environment=to_pascal_case(project.get('environment')) or None,
        ServiceRole=project.get('serviceRole'),
        TimeoutInMinutes=project.get('timeoutInMinutes'),
        QueuedTimeoutInMinutes=project.get('queuedTimeoutInMinutes'),
        EncryptionKey=project.get('encryptionKey'),
        Tags=to_pascal_case(tags) or None,
        VpcConfig=project.get('vpcConfig') or None,
        BadgeEnabled=project.get('badge', {}).get('badgeEnabled'),
        LogsConfig=project.get('logsConfig') or None,
        FileSystemLocations=project.get('fileSystemLocations') or None,
        BuildBatchConfig=project.get('buildBatchConfig') or None,
        ConcurrentBuildLimit=project.get('concurrentBuildLimit'),
        ResourceAccessRole=project.get('resourceAccessRole'),
        Triggers=None,
        Visibility=project.get('projectVisibility'),
    )
    properties = { k: v for k, v in properties.items() if v is not None }
    data = {
        logical_name: dict(
            Type='AWS::CodeBuild::Project',
            DeletionPolicy='Retain',
            Properties=properties,
        ),
    }
    print()
    dump_yaml(data)
    print()



import os
import posixpath
import re
from configparser import ConfigParser
from os.path import expanduser

from raft import task, Collection
from ...base.utils import notice, notice_end
from ..base import AwsTask


codeartifact_collection = Collection()


@task(klass=AwsTask)
def login_to_repo(
        ctx, name=None, domain=None, owner=None,
        repo_format='pypi', profile=None,
        pip=False, twine=False,
        region=None, session=None, **kwargs):
    """
    logs into the specified codeartifact pypi repository and configures pip

    config sample:

    ```yaml
    codeartifact:
      default_repo: myrepo
      myrepo:
        domain: mydomain
        owner: 012345678901
        format: npm
        region: us-west-1
    ```
    """
    from boto3.session import Session
    from convocations.base.utils import get_context_value as ctx_value
    name = name or ctx_value(ctx, 'codeartifact.default_repo')
    prefix = f'codeartifact.{name}'
    domain = domain or ctx_value(ctx, f'{prefix}.domain')
    owner = owner or ctx_value(ctx, f'{prefix}.owner')
    repo_format = repo_format or ctx_value(ctx, f'{prefix}.format')
    repo_format = repo_format or 'pypi'
    region = region or ctx_value(ctx, f'{prefix}.region')
    login_session = session
    if region and session.region_name != region:
        login_session = Session(
            profile_name=session.profile_name,
            region_name=region)
    client = login_session.client('codeartifact')
    notice(f'getting {name} endpoint')
    params = dict(
        domain=domain,
        repository=name,
        format=repo_format,
    )
    if owner:
        params['domainOwner'] = f'{owner}'
    response = client.get_repository_endpoint(**params)
    endpoint = response['repositoryEndpoint']
    notice_end()
    notice('getting authorization token')
    del params['repository']
    del params['format']
    params['durationSeconds'] = 43200
    response = client.get_authorization_token(**params)
    token = response['authorizationToken']
    notice_end()
    if pip:
        notice('configuring pip')
        url = f'{endpoint[:8]}aws:{token}@{endpoint[8:]}'
        url = posixpath.join(url, 'simple')
        ctx.run('pip config set global.timeout 10', hide='both')
        ctx.run(f'pip config set global.extra-index-url {url}', hide='both')
        notice_end()
    if twine:
        config = ConfigParser()
        twine_config_path = expanduser('~/.pypirc')
        notice(f'reading {twine_config_path}')
        if os.path.exists(twine_config_path):
            config.read(twine_config_path)
            notice_end()
        else:
            notice_end(False)
        notice('configuring twine')
        if not config.has_section('distutils'):
            config.add_section('distutils')
        index_servers = config['distutils'].get('index-servers') or ''
        index_servers = index_servers.split('\n')
        index_servers = [ x.strip() for x in index_servers ]
        index_servers.append(name)
        index_servers = '\n'.join(index_servers)
        config['distutils']['index-servers'] = index_servers
        if not config.has_section(name):
            config.add_section(name)
        config[name]['username'] = 'aws'
        config[name]['password'] = token
        config[name]['repository'] = endpoint
        with open(twine_config_path, 'w') as f:
            config.write(f)
        notice_end()


codeartifact_collection.add_task(login_to_repo)




import os
import sys
from raft import task

from convocations.base.utils import get_context_value


@task
def global_python(ctx, pythons=None):
    """
    finds a global python interpreter in order of preference

    example config:

    ```yaml
    py:
      pythons:
        - /path/to/python3.8
        - /path/to/python2.7
    ```
    :param ctx:
    :param pythons:
    :return:
    """
    possible = pythons or get_context_value(ctx, 'py.pythons') or [
        'python3.10',
        'python3.9',
        'python3.8',
        'python3.7',
        'python3.6',
    ]
    if sys.platform != 'win32':
        for x in possible:
            result = ctx.run(f'which {x}', warn=True)
            if result.ok:
                return result.stdout.strip()
    else:
        for x in possible:
            st = x.replace('.', '')
            st = os.path.join('/', st, 'python.exe')
            if os.path.exists(st):
                return st
        result = ctx.run('(Get-Command python).source', warn=True)
        if result.ok:
            return result.stdout.strip()
    return None


@task
def setup(ctx, name=None):
    """
    creates the virtual environment and installs prerequisites
    example config:

    ```yaml
    py:
      venv_name: .venv
      windows_venv_name: .wenv
    ```
    """
    from ..aws.codebuild.detect import is_codebuild
    if is_codebuild():
        from ..aws.codebuild import setup as cb_setup
        cb_setup(ctx)
        return

    default_name = '.'
    if sys.platform in ('linux', 'darwin'):
        default_name = get_context_value(ctx, 'py.venv_name', '.')
    elif sys.platform == 'win32':
        default_name = get_context_value(ctx, 'py.windows_venv_name', '.wenv')
    name = name or default_name
    expected_python = os.path.join(name, 'bin', 'python')
    if not os.path.exists(expected_python):
        python = global_python(ctx)
        ctx.run(f'{python} -m venv {name}')
    if os.path.exists(expected_python):
        python = expected_python
        pip = f'{python} -m pip install'
        result = ctx.run(f'{pip} -U wheel', pty=True)
        if result.exited:
            return
        result = ctx.run(f'{pip} -U raft', pty=True)
        if result.exited:
            return
        result = ctx.run(f'{pip} -r requirements.txt', pty=True)
        if result.exited:
            return
        if os.path.exists('setup.py'):
            result = ctx.run(f'{pip} -e .', pty=True)
            if result.exited:
                return
    else:
        print('no suitable python version found')



from typing import List, Optional
from cryptography import x509
from cryptography.x509.name import NameAttribute, ObjectIdentifier
from cryptography.hazmat.backends import default_backend
from cryptography.x509.extensions import ExtensionNotFound
from raft import task
from ..base.utils import notice, notice_end
from ..base.utils import print_table


def read_chain(filename: str) -> List[bytes]:
    with open(filename, 'r') as f:
        data = f.read()
    data = data.split('-----END CERTIFICATE-----')
    data = [ x.strip() for x in data ]
    return [ f'{x}\n-----END CERTIFICATE-----\n'.encode('ascii') for x in filter(None, data) ]


def common_name(x: x509.name.Name) -> Optional[str]:
    oid: ObjectIdentifier = ObjectIdentifier('2.5.4.3')
    attributes: List[NameAttribute] = x.get_attributes_for_oid(oid)
    if not attributes:
        return None
    return attributes[0].value


def yield_bytes(st):
    for x in range(0, len(st), 2):
        yield st[x:x + 2]


def get_extension(x, key: str) -> x509.extensions.Extension:
    oid: ObjectIdentifier = ObjectIdentifier(key)
    e: x509.extensions.Extension = x.extensions.get_extension_for_oid(oid)
    return e


def subject_fingerprint(x) -> Optional[str]:
    e = get_extension(x, '2.5.29.14')
    key_id: x509.extensions.SubjectKeyIdentifier = e.value
    return ':'.join(yield_bytes(key_id.digest.hex()))


def issuer_fingerprint(x) -> Optional[str]:
    try:
        e = get_extension(x, '2.5.29.35')
        key_id: x509.extensions.AuthorityKeyIdentifier = e.value
        return ':'.join(yield_bytes(key_id.key_identifier.hex()))
    except ExtensionNotFound:
        return ''


@task
def validate_chain(ctx, filename):
    """
    validates a bundled crt file
    """
    backend = default_backend()
    certs = read_chain(filename)
    values, values2 = [], []
    for cert in certs:
        cert = x509.load_pem_x509_certificate(cert, backend)
        subject = common_name(cert.subject)
        issuer = common_name(cert.issuer)
        subject_key = subject_fingerprint(cert)
        issuer_key = issuer_fingerprint(cert)
        expiration = cert.not_valid_after.strftime('%Y-%m-%d')
        values.append([ subject, issuer, expiration ])
        values2.append([ subject_key, issuer_key ])
    notice('validating issuer chain')
    valid = True
    for x in range(1, len(values2)):
        if values2[x][0] != values2[x - 1][-1]:
            valid = False
    if valid:
        notice_end()
    else:
        notice_end('invalid')
    print()
    print_table([ 'subject', 'issuer', 'expiration', ], values)
    print()
    print_table([ 'subject fingerprint', 'issuer fingerprint', ], values2)



from raft.tasks import task
from ...base.utils import dump_yaml
from ..base import AwsTask
from ..base import yielder


@task(klass=AwsTask)
def ec2_events(ctx, session=None, **kwargs):
    """
    shows all health events for ec2 instances that are
    in status open
    """
    health = session.client('health')
    fn = yielder(health, 'describe_events', session=session)
    events = list(fn)
    dump_yaml(events, quiet=False)



from raft import task
from .base import XmlApiTask, XmlApi
from ..base.utils import notice, notice_end


@task(klass=XmlApiTask)
def remove_hip_profiles(ctx, host=None, config=None,
                        device='localhost.localdomain',
                        vsys='vsys1', **kwargs):
    """
    if we see weird hip profiles in the rulebase, this convocation removes them
    """
    from lxml import etree
    notice(f'connecting to {host}')
    api = kwargs.get('api')
    api = api or XmlApi(host, filename=config)
    doc = api.get('config', 'show')
    notice_end()
    notice('searching rules')
    xpath = (
        f"devices/entry[@name='{device}']/vsys/"
        f"entry[@name='{vsys}']/rulebase/security"
    )
    doc_xpath = f'{xpath}/rules'
    xpath = (
        f"/config/devices/entry[@name='{device}']/"
        f"vsys/entry[@name='{vsys}']/rulebase/security"
    )
    rulebase = doc.xpath(doc_xpath)
    rules = rulebase[0]
    entries = rules.xpath('entry')
    notice_end(f'{len(entries)}')
    notice('looking for hip profiles')
    hip_profiles = rules.xpath('entry/hip-profiles')

    for x in hip_profiles:
        x.getparent().remove(x)
    notice_end(f'{len(hip_profiles)}')
    for x in entries:
        notice(f"updating {x.attrib['name']}")
        entry_xpath = f"{xpath}/rules/entry[@name='{x.attrib['name']}']"
        inner_xml = etree.tostring(x).decode('utf-8')
        response = api.post('config', 'edit', xpath=entry_xpath, element=inner_xml)
        notice_end(etree.tostring(response).decode('utf-8'))



from raft.collection import Collection
from .codebuild import codebuild_collection
from .codebuild import global_collection as global_codebuild
from . import codebuild, rds
from . import codedeploy
from . import ec2
from . import elb
from . import iam
from . import ami_helpers
from . import open_id
from .keypairs import new_key_pair, show_public_key
from .cfn import import_instance
from .cfn import import_route_table
from .cfn import import_security_group
from .cfn import import_subnet
from .cfn import import_vpc
from .cfn import import_ecs_service
from .cfn import import_code_pipeline
from .cfn import stack_commands
from . import mgn
from . import transit_gateway
from . import ram
from .ssm import runners as ssm_runners
from .ssm.runners import ssm_parameters
from . import s3
from . import organizations
from . import route53
from . import vpc_peering
from .route53 import zone_parser
from . import reservations
from .ec2 import routing
from .ec2 import ssh
from .ec2 import networking as ec2_networking
from .ip_ranges import ip_ranges
from .health import ec2_events
from . import efs


ns = Collection()
ns.add_collection(codebuild_collection, 'codebuild')
ns.add_tasks(codebuild.build)
ns.add_tasks(codedeploy)
ns.add_tasks(ec2)
ns.add_tasks(ec2_networking)
ns.add_tasks(elb)
ns.add_tasks(iam)
ns.add_tasks(import_instance)
ns.add_tasks(import_route_table)
ns.add_tasks(import_security_group)
ns.add_tasks(import_subnet)
ns.add_tasks(import_vpc)
ns.add_tasks(import_ecs_service)
ns.add_tasks(mgn)
ns.add_tasks(stack_commands)
ns.add_tasks(transit_gateway)
ns.add_tasks(open_id)
ns.add_tasks(ami_helpers)
ns.add_task(new_key_pair)
ns.add_task(show_public_key)
ns.add_task(ssm_runners.run_posh)
ns.add_tasks(ram)
ns.add_tasks(s3)
ns.add_tasks(organizations)
ns.add_tasks(route53)
ns.add_tasks(zone_parser)
ns.add_tasks(vpc_peering)
ns.add_tasks(import_code_pipeline)
ns.add_tasks(rds)
ns.add_tasks(reservations)
ns.add_tasks(routing)
ns.add_tasks(ssh)
ns.add_task(ip_ranges)
ns.add_task(ec2_events)
ns.add_task(ssm_parameters)
ns.add_tasks(efs)



import posixpath
from datetime import datetime
from datetime import timedelta

from raft import task

from ...base.utils import notice, notice_end, print_table
from ..base import AwsTask


@task(klass=AwsTask, help={
    'name': 'the application name in code deploy',
    'group_name': 'the deployment group name; if left unspecified, we use the app_name',
    'lookback_window': 'the number of days in which we will look for deploys',
})
def recent_deployments(ctx, name, group_name=None,
                       lookback_window=30,
                       session=None, profile=None, **kwargs):
    """
    shows the most recent deploys and their statuses
    """
    client = session.client('codedeploy')
    notice('getting deployment group')
    result = client.batch_get_deployment_groups(
        applicationName=name,
        deploymentGroupNames=[group_name or name],
    )
    group_id = None
    if result:
        group_id = result['deploymentGroupsInfo'][0]['deploymentGroupId']
    notice_end(group_id)
    notice('getting deployments')
    end_date = datetime.today() + timedelta(days=1)
    start_date = end_date - timedelta(days=lookback_window + 1)
    result = client.list_deployments(
        applicationName=name,
        deploymentGroupName=group_name or name,
        createTimeRange=dict(
            start=start_date,
        )
    )
    notice_end(len(result['deployments']))
    values = [ ]
    deployment_ids = result['deployments'][-25:]
    result = client.batch_get_deployments(deploymentIds=deployment_ids)
    deployments = result['deploymentsInfo']
    for x in deployments:
        row = [
            x['deploymentId'] or '',
            x['status'].lower(),
            x['createTime'].isoformat(' ', 'seconds') if 'createTime' in x else '',
            x['startTime'].isoformat(' ', 'seconds') if 'startTime' in x else '',
            x['completeTime'].isoformat(' ', 'seconds') if 'completeTime' in x else '',
        ]
        values.append(row)
    header = [ 'id', 'status', 'created', 'started', 'ended']
    print_table(header, values)


@task(klass=AwsTask, help={
    'name': 'the application name in code deploy',
    'group_name': 'the deployment group name; if left unspecified, we use the app_name',
    's3_location': 'the s3 url for the deployment bundle',
})
def code_deploy(ctx, name, s3_location, group_name=None,
           session=None, profile=None, **kwargs):
    """
    kicks off the deploy name for the specified app and deployment group
    """
    group_name = group_name or name
    if s3_location.startswith('s3://'):
        s3_location = s3_location[5:]
    npos = s3_location.index('/')
    bucket = s3_location[:npos]
    key = s3_location[npos + 1:]
    _, ext = posixpath.splitext(key)
    ext = ext[1:].lower()
    types = dict(
        json='JSON',
        yml='YAML',
        yaml='YAML',
        zip='zip',
        tar='tar',
        tgz='tgz',
    )
    client = session.client('codedeploy')
    notice('starting deploy')
    client.create_deployment(
        applicationName=name,
        deploymentGroupName=group_name or name,
        revision=dict(
            revisionType='S3',
            s3Location=dict(
                bucket=bucket,
                key=key,
                bundleType=types[ext],
            )
        )
    )
    notice_end()




from os.path import abspath, dirname, join

from raft.collection import Collection
from convocations.base.utils import load_yaml


ns = Collection()

filename = dirname(dirname(abspath(__file__)))
filename = join(filename, 'conf', 'convocations.yml')
conf = load_yaml(filename)
ns.configure(conf)



from raft import task
from ...base.utils import print_table, notice, notice_end, dump_yaml
from ..base import AwsTask


@task(klass=AwsTask)
def subnet_to_yaml(
        ctx, logical_name=None, subnet_id=None,
        session=None, quiet=False, **kwargs):
    """
    generates the cloudformation yaml for a security group
    """
    ec2 = session.resource('ec2')
    subnet = ec2.Subnet(subnet_id)
    properties = dict(
        AssignIpv6AddressOnCreation=subnet.assign_ipv6_address_on_creation,
        AvailabilityZone=subnet.availability_zone,
        CidrBlock=subnet.cidr_block,
        EnableDns64=subnet.enable_dns64,
        Ipv6CidrBlock=subnet.ipv6_cidr_block_association_set[0] if subnet.ipv6_cidr_block_association_set else None,
        Ipv6Native=subnet.ipv6_native,
        MapPublicIpOnLaunch=subnet.map_public_ip_on_launch,
        OutpostArn=subnet.outpost_arn,
        PrivateDnsNameOptionsOnLaunch=subnet.private_dns_name_options_on_launch,
        Tags=subnet.tags,
        VpcId=subnet.vpc_id,
    )
    properties = { k: v for k, v in properties.items() if v is not None }
    if 'Ipv6CidrBlock' not in properties:
        del properties['AssignIpv6AddressOnCreation']
    data = {
        logical_name: dict(
            Type='AWS::EC2::Subnet',
            DeletionPolicy='Retain',
            Properties=properties,
        ),
    }
    if not quiet:
        dump_yaml(data, quiet=False)
    return data


@task(klass=AwsTask)
def import_subnet(ctx, logical_name, subnet_id, filename,
                  skip_review=False, session=None, prefix=None,
                  **kwargs):
    """
    creates a yaml template for a security group and imports it into cloudformation

    if skip_review is set, we will immediately try to execute the change set
      in cloudformation
    """
    from .import_resource_to_cfn import import_resource_to_cfn
    data = subnet_to_yaml(
        ctx, logical_name, subnet_id, session=session, quiet=True)
    resources_to_import = [dict(
        ResourceType='AWS::EC2::Subnet',
        LogicalResourceId=list(data.keys())[0],
        ResourceIdentifier=dict(SubnetId=subnet_id),
    )]
    import_resource_to_cfn(
        data, resources_to_import,
        filename, None, skip_review, session, prefix)


@task(klass=AwsTask)
def subnets(ctx, name=None, routes=False, session=None, **kwargs):
    from ..ec2 import get_tag
    from ..base import yielder
    notice('loading subnets')
    ec2 = session.client('ec2')
    rg_subnets = list(yielder(ec2, 'describe_subnets', session=session))
    notice_end()
    header = [ 'id', 'name', 'cidr', 'cidr6', 'az', ]
    if routes:
        notice('loading route tables')
        rg_tables = list(yielder(ec2, 'describe_route_tables', session=session))
        notice_end()
        mp_subnets = { x['SubnetId']: x for x in rg_subnets }
        for x in rg_tables:
            table_id = x['RouteTableId']
            table_name = get_tag(x, 'Name')
            for association in x['Associations']:
                subnet_id = association.get('SubnetId')
                if not subnet_id:
                    continue
                subnet = mp_subnets[subnet_id]
                subnet['route_table_id'] = table_id
                subnet['route_table'] = table_name
        header += [ 'route_table_id', 'route_table' ]
    rows = []
    name = (name or '').lower()
    for x in rg_subnets:
        subnet_name = get_tag(x, 'Name') or ''
        subnet_name = subnet_name.lower()
        if name not in subnet_name:
            continue
        row = [
            x['SubnetId'], get_tag(x, 'name') or '', x.get('CidrBlock'),
        ]
        ipv6 = x.get('Ipv6CidrBlockAssociationSet') or [{}]
        row.append(ipv6[0].get('Ipv6CidrBlock', ''))
        row.append(x.get('AvailabilityZone'))
        if routes:
            row.append(x.get('route_table_id', ''))
            row.append(x.get('route_table', ''))
        rows.append(row)
    rows.sort(key=lambda lx: lx[1])
    print_table(header, rows)



import warnings

from raft.context import Context
from ..aws.base import AwsTask
from ..base.utils import get_context_value

try:
    from urllib3.exceptions import InsecureRequestWarning
    import requests
    import yaml
    from murmuration import kms_wrapped
    from lxml import etree
except ImportError:
    pass


class XmlApiTask(AwsTask):
    def __call__(self, *args, **kwargs):
        ctx = args[0]
        host = kwargs.pop('host', None)
        bucket = kwargs.pop('bucket', None)
        if isinstance(ctx, Context):
            kwargs['host'] = host or get_context_value(ctx, 'palo_alto.host')
            kwargs['bucket'] = bucket or get_context_value(ctx, 'palo_alto.bucket')
            kwargs['username'] = get_context_value(ctx, 'palo_alto.username')
            kwargs['password'] = get_context_value(ctx, 'palo_alto.password')
            kwargs['peer'] = get_context_value(ctx, 'palo_alto.peer')
        super().__call__(*args, **kwargs)


class XmlApi:
    def __init__(self, host, username=None, password=None, profile=None, **kwargs):
        self.session = requests.Session()
        self.session.verify = False
        self.host = host
        self.base_url = f'https://{self.host}/api/'
        self.generate(host, username, password, profile, **kwargs)

    @classmethod
    def resolve_creds(cls, username, password, session=None):
        from ..aws.base import get_parameter
        if username.startswith('/'):
            username = get_parameter(session, username)
        if password.startswith('/'):
            password = get_parameter(session, password, True)
        return username, password

    def generate(self, host, username=None, password=None, profile=None, **kwargs):
        """
        generates an api key using the rest api with the palo alto
        """
        session = kwargs.get('session')
        filename = kwargs.get('filename')
        if filename:
            with open(filename, 'r') as f:
                conf = yaml.load(f, Loader=yaml.SafeLoader)
            api_conf = conf.get('api') or {}
            username = username or api_conf.get('username')
            password = password or api_conf.get('password')
            profile = profile or conf.get('profile')
        username, password = self.resolve_creds(username, password, session)
        if not password:
            try:
                password = kms_wrapped.decrypt(password, profile=profile)
            except:
                pass

        doc = self.post('keygen', None, user=username, password=password)
        api_key = doc.text
        self.session.headers = {
            'X-PAN-KEY': api_key,
        }

    def raw(self, method, params=None, data=None, files=None):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=InsecureRequestWarning)
            response = self.session.request(
                method, self.base_url,
                params=params, data=data, files=files)
            return response

    def get(self, type_, action=None, **kwargs):
        params = {}
        params.update(kwargs)
        params['type'] = type_
        if action:
            params['action'] = action
        response = self.raw('get', params)
        return self.handle_response(response)

    @classmethod
    def handle_response(cls, response):
        doc = etree.fromstring(response.content)
        if doc.tag == 'response':
            stuff = doc.xpath('/response/result')
            if stuff:
                stuff = stuff[0]
            for x in stuff:
                return x
        return doc

    def post(self, type_, action, **kwargs):
        params = {}
        params.update(kwargs)
        params['type'] = type_
        if action:
            params['action'] = action
        response = self.raw('post', data=params)
        return self.handle_response(response)

    def import_cert(self, name, category,
                    certificate_format='pem',
                    filename=None, content=None, **kwargs):
        params = {}
        params.update(kwargs)
        params['type'] = 'import'
        params['category'] = category
        params['certificate-name'] = name
        params['format'] = certificate_format
        files = {
            'certFile': content,
        }
        if filename and not content:
            with open(filename, 'r') as f:
                files['certFile'] = f.read()
        response = self.raw('post', data=params, files=files or None)
        return self.handle_response(response)

    def import_pfx(self, name, content, passphrase, **kwargs):
        params = {}
        params.update(kwargs)
        params['type'] = 'import'
        params['category'] = 'keypair'
        params['certificate-name'] = name
        params['format'] = 'pkcs12'
        params['passphrase'] = passphrase
        files = {
            'file': (name, content, 'application/x-pkcs12'),
        }
        response = self.raw('post', data=params, files=files or None)
        return self.handle_response(response)

    def import_pem(self, name, cert_content, key_content=None, passphrase=None,
                   **kwargs):
        params = {}
        params.update(kwargs)
        params['type'] = 'import'
        params['category'] = 'keypair' if key_content else 'certificate'
        params['certificate-name'] = name
        params['format'] = 'pem'
        if passphrase:
            params['passphrase'] = passphrase
        mime_type = 'application/x-pem-file'
        files = {
            'certFile': (f'{name}.crt', cert_content, mime_type),
        }
        if key_content:
            files['keyFile'] = (f'{name}.key', key_content, mime_type)
        response = self.raw('post', data=params, files=files or None)
        return self.handle_response(response)

    def commit(self, description):
        params = {
            'description': description,
            'cmd': '<commit></commit>'
        }
        params['type'] = 'commit'
        response = self.raw('post', data=params)
        return self.handle_response(response)



import string
import random
from raft.tasks import task
from .base import AzureTask


@task(klass=AzureTask)
def update_ldaps_cert(ctx, certificate, key,
                      domain=None, resource_group=None,
                      client_id=None, client_secret=None,
                      subscription_id=None, creds=None, **kwargs):
    """
    updates ldaps cet for the specified domain
    use https://learn.microsoft.com/en-us/azure/templates/microsoft.aad/domainservices?pivots=deployment-language-bicep
    for the syntax of the update object
    """
    from convocations.base.utils import get_context_value
    from convocations.base.utils import notice, notice_end
    from azure.mgmt.resource.resources.models import GenericResource
    from azure.mgmt.resource.resources import ResourceManagementClient
    from cryptography.x509 import load_pem_x509_certificates
    from cryptography.hazmat.primitives.serialization import BestAvailableEncryption
    from cryptography.hazmat.primitives.serialization import \
        load_pem_private_key
    from cryptography.hazmat.primitives.serialization.pkcs12 import \
        serialize_key_and_certificates
    from base64 import b64encode
    notice('building pfx')
    allowed = string.ascii_uppercase + string.ascii_lowercase + string.digits
    st_password = ''.join([ random.choice(allowed) for _ in range(16) ])
    with open(certificate, 'rb') as f:
        certs = load_pem_x509_certificates(f.read())
    with open(key, 'rb') as f:
        key = load_pem_private_key(f.read(), password=None)
    en_password = BestAvailableEncryption(st_password.encode('utf-8'))
    p12 = serialize_key_and_certificates(
        None,
        key,
        certs[0],
        certs[1:] or None,
        en_password
    )
    notice_end()
    notice('updating ldaps cert')
    subscription_id = subscription_id or get_context_value(ctx, 'azure.subscription_id')
    domain = domain or get_context_value(ctx, 'azure.domain')
    resource_group = resource_group or get_context_value(ctx, 'azure.resource_group')
    client = ResourceManagementClient(credential=creds, subscription_id=subscription_id)
    parameters = GenericResource()
    parameters.properties = {
        'ldapsSettings': {
            'ldaps': 'Enabled',
            'externalAccess': 'Disabled',
            'pfxCertificate': b64encode(p12).decode('utf-8'),
            'pfxCertificatePassword': st_password,
        }
    }
    response = client.resources.begin_update(
        resource_name=domain,
        resource_group_name=resource_group,
        resource_provider_namespace='Microsoft.AAD',
        parent_resource_path='',
        resource_type='DomainServices',
        api_version='2022-12-01',
        parameters=parameters,
    )
    for x in range(30):
        response.wait(5)
        if response.done():
            break
    notice_end()



import requests
from raft import task
from convocations.base.utils import print_table
from convocations.base.utils import notice, notice_end


@task
def ip_ranges(ctx, region='us-east-2', service='amazon'):
    """
    filters aws public ip ranges by region and service
    """
    notice('getting ip ranges')
    url = 'https://ip-ranges.amazonaws.com/ip-ranges.json'
    response = requests.get(url, timeout=10)
    data = response.json()
    prefixes = data['prefixes']
    notice_end(f'{len(prefixes)}')
    notice('filtering for region / service')
    i = 0
    rows = []
    for x in prefixes:
        if x['region'] == region and x['service'].lower() == service.lower():
            rows.append([ x['ip_prefix'] ])
            i += 1
    notice_end(f'{i}')
    print_table([ 'cidr' ], rows)



#!/usr/bin/env python

# Support setuptools only, distutils has a divergent and more annoying API and
# few folks will lack setuptools.
from setuptools import setup, find_packages
from importlib.resources import open_text
# Version info -- read without importing
_locals = {}

version = '1.7'

# PyYAML ships a split Python 2/3 codebase. Unfortunately, some pip versions
# attempt to interpret both halves of PyYAML, yielding SyntaxErrors. Thus, we
# exclude whichever appears inappropriate for the installing interpreter.
exclude = ["*.yaml2", 'test']

# Frankenstein long_description: version-specific changelog note + README
with open('README.md') as f:
    long_description = f.read()

extras = {}
all_extras = set()
for x in [ 'dev', 'checkpoint', 'onelogin', 'certs', 'aws', 'sharepoint',
           'palo_alto', 'tundra', 'airflow', 'azure', ]:
    filename = f'{x}.txt'
    with open(filename, 'r') as f:
        st = f.read()
    rg = st.split('\n')
    extras[x] = [ lx for lx in rg if lx ]
    if x != 'dev':
        all_extras |= set(rg)
if all_extras:
    all_extras = list(filter(None, all_extras))
    all_extras.sort()
    extras['all'] = all_extras
print(extras)

setup(
    name='convocations',
    version=version,
    description='convocations',
    license='BSD',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='2ps',
    author_email='pshingavi@gmail.com',
    url='https://github.com/crazy-penguins/convocations',
    package_dir={ '': '.', },
    packages=find_packages(
        where='.',
        exclude=['test', ]),
    package_data={
        'convocations.pypi': [ 'templates/**/*', ],
        'convocations.tundra': [ 'templates/**/*', ],
        'convocations.docker': [ 'templates/**/*', ],
    },
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=[
        'jinja2',
        'pyyaml',
        'termcolor',
        'colorama',
        'raft',
        'requests',
        'requests_oauthlib',
        'murmuration>=0.8',
        'waddle>=1.2',
        'halo',
        'python-dateutil',
        'pytz',
    ],
    entry_points={
       'console_scripts': [
           'pypi = convocations.pypi.program:program.run',
           'sharepoint = convocations.sharepoint.program:program.run',
       ]
    },
    extras_require=extras,
)



"""
front-end client tasks for local development in tundra
"""
from raft import task


@task(default=True)
def client(ctx):
    """ runs the client watch """
    ctx.run('rm -rf client/static/*')
    ctx.run('. nenv/bin/activate && cd client && npm run client')


@task
def build(ctx):
    """ builds the client """
    ctx.run('rm -rf client/static/*')
    ctx.run('make setup_client')
    ctx.run('. nenv/bin/activate && cd client && npm run build')


@task
def build_docker(ctx):
    """ builds the client using a docker-compose cmd """
    ctx.run('rm -rf client/static/*')
    ctx.run('docker-compose -f client/docker-compose.yml up')



import re
from raft.tasks import task
from ..base import AwsTask, yielder
from ...base.utils import print_table, notice, notice_end


@task(klass=AwsTask)
def run_posh(ctx, name, filename, session=None, **kwargs):
    """
    runs a command on the specified windows instance using ssm run command
    :param ctx: the raft / convoke context
    :param str name: the name or instance id of the ec2 instance
    :param str filename: the /path/to the .ps1 script that you want to run
    :param session: leave this blank
    """
    from ..ec2 import yield_instances
    from ..base import get_tag
    notice(f'loading {filename}')
    with open(filename, 'rt') as f:
        commands = f.readlines()
    instances = list(yield_instances(name, session))
    instance_ids = [ x['InstanceId'] for x in instances ]
    header = [ 'instance_id', 'name' ]
    rows = []
    for x in instances:
        rows.append([ x['InstanceId'], get_tag(x, 'name') ])
    print()
    print_table(header, rows)
    print()
    notice(f'running command on {len(rows)} instances')
    ssm = session.client('ssm')
    result = ssm.send_command(
        InstanceIds=instance_ids,
        DocumentName='AWS-RunPowerShellScript',
        Parameters=dict(commands=commands)
    )
    result = result['Command']
    command_id = result['CommandId']
    notice_end(command_id)
    waiter = ssm.get_waiter('command_executed')
    for instance_id in instance_ids:
        notice(f'waiting for {instance_id}')
        waiter.wait(
            CommandId=command_id,
            InstanceId=instance_id,
            WaiterConfig=dict(Delay=15))
        notice_end()
        x = ssm.get_command_invocation(CommandId=command_id, InstanceId=instance_id)
        print(x['StandardOutputContent'])
        print(x['StandardErrorContent'])


@task(klass=AwsTask)
def ssm_parameters(ctx, name=None, regexp=None, session=None, **kwargs):
    """
    finds all parameters with `name` contained in the name of the parameter
    and also prints the values
    """
    notice('finding parameters')
    client = session.client('ssm')
    if regexp:
        matcher = re.compile(regexp, re.IGNORECASE)
        name = ''
    name = name or ''
    params = {
        'PaginationConfig': {'PageSize': 50}
    }
    if name:
        params['ParameterFilters'] = [{
            'Key': 'Name',
            'Option': 'Contains',
            'Values': [ name, ],
        }]
    fn = yielder(client, 'describe_parameters', session=session, **params)
    parameters = list(fn)
    notice_end()
    header = [ 'name', 'value' ]
    rows = []
    for i, x in enumerate(parameters, 1):
        if i % 10 == 1:
            if i > 1:
                notice_end()
            n_end = min(i + 9, len(parameters))
            notice(f'getting values {i}-{n_end}/{len(parameters)}')
        name = x['Name']
        if regexp and not matcher.search(name):
            continue
        t = x['Type']
        response = client.get_parameter(
            Name=name,
            WithDecryption=t == 'SecureString',
        )
        value = response['Parameter']['Value']
        rows.append([ name, value ])
    rows.sort(key=lambda lx: lx[0])
    notice_end()
    print_table(header, rows)



from raft import Collection
from ..aws.codebuild import build
from ..aws.codeartifact import login_to_repo
from ..aws.open_id import sso_login
from . import deploy
from . import codebuild


tundra_collection = Collection()
tundra_collection.add_tasks(build)
tundra_collection.add_tasks(login_to_repo)
tundra_collection.add_task(sso_login)
tundra_collection.add_collection(Collection.from_module(deploy), 'deploy')
tundra_collection.add_collection(Collection.from_module(codebuild), 'codebuild')



from raft.tasks import task
from .base import AzureTask


@task(klass=AzureTask)
def subscriptions(ctx, tenant_id=None, client_id=None, client_secret=None, creds=None, **kwargs):
    """
    prints a list of subscriptions available in this account
    """
    from azure.mgmt.resource.subscriptions import SubscriptionClient
    from ..base.utils import print_table
    client = SubscriptionClient(credential=creds)
    header = [ 'id', 'name' ]
    rows = []
    for x in client.subscriptions.list():
        rows.append([
            x.subscription_id,
            x.display_name,
        ])
    print_table(header, rows)





from raft import Collection
from .save_config import save_config
from .save_config import export_config
from .upgrade import upgrade
from .weird import remove_hip_profiles
from .certificates import install_pfx
from .certificates import install_pem


panos = Collection(
    save_config,
    export_config,
    upgrade,
    remove_hip_profiles,
    install_pfx,
    install_pem,
)



import os
from raft.context import Context
from raft.tasks import Task


def connect(host):
    from pwinput import pwinput
    from cpapi import APIClient, APIClientArgs
    args = APIClientArgs(server=host)
    client = APIClient(args)
    default = os.environ.get('USER', os.environ.get('USERNAME'))
    username = input(f'checkpoint username (default: {default}): ') or default
    password = pwinput('checkpoint password: ')
    client.login(username, password, read_only=True)
    return client


class CheckpointTask(Task):
    def __call__(self, *args, **kwargs):
        client = kwargs.get('client')
        if client:
            return super().__call__(*args, **kwargs)

        host = kwargs.get('host')
        ctx = args[0]
        if host is None and isinstance(ctx, Context):
            try:
                host = ctx.checkpoint.host
            except AttributeError:
                pass
        client = connect(host)
        kwargs['client'] = client
        result = super().__call__(*args, **kwargs)
        client.close_connection()
        return result




from raft import task
from .base import XmlApiTask, XmlApi
from ..base.utils import notice, notice_end

try:
    from lxml import etree
except ImportError:
    pass


@task(klass=XmlApiTask)
def save_config(ctx, host=None, bucket=None, username=None, password=None,
                api=None, session=None, **kwargs):
    """
    saves off the running config from the specified hosts.
    if bucket is specified, the config is saved as the bootstrap
    config in the specified bucket
    """
    notice('getting config')
    api = api or XmlApi(host, username, password, session=session)
    doc = api.get('config', 'show')
    config = etree.tostring(doc, pretty_print=True)
    notice_end()
    if not bucket:
        print(config.decode('utf-8'))
    else:
        key = f'{host}/config.xml'
        notice(f'upload to s3://{bucket}/{key}')
        s3 = session.client('s3')
        s3.put_object(
            Bucket=bucket,
            Key=key,
            Body=etree.tostring(doc, pretty_print=True))
        notice_end()


@task(klass=XmlApiTask)
def export_config(ctx, host=None, bucket=None, session=None, username=None,
                  password=None, api=None, **kwargs):
    """
    exports the running config from the specified hosts.
    if bucket is specified, the config is saved as the bootstrap
    config in the specified bucket
    """
    notice('exporting config')
    api = api or XmlApi(host, username, password, session.profile_name)
    doc = api.get('export', category='configuration', )
    config = etree.tostring(doc, pretty_print=True)
    notice_end()
    if not bucket:
        print(config.decode('utf-8'))
    else:
        key = f'{host}/config.xml'
        notice(f'upload to s3://{bucket}/{key}')
        s3 = session.client('s3')
        s3.put_object(
            Bucket=bucket,
            Key=key,
            Body=etree.tostring(doc, pretty_print=True))
        notice_end()
